import hcat.lib.functional
from hcat.lib.utils import calculate_indexes, load, cochlea_to_xml, correct_pixel_size, scale_to_hair_cell_diameter
from hcat.lib.cell import Cell
from hcat.lib.cochlea import Cochlea
from hcat.backends.detection import FasterRCNN_from_url

import torch
from torch import Tensor
from tqdm import tqdm
from itertools import product
import numpy as np
from hcat.lib.explore_lif import get_xml
import torchvision.ops
import skimage.io as io

import os.path
from typing import List, Dict


# DOCUMENTED

def _detect(f: str, curve_path: str = None, cell_detection_threshold: float = 0.86, dtype=None,
            nms_threshold: float = 0.2, save_xml=False, save_fig=False, pixel_size=None, cell_diameter=None):
    """
    2D hair cell detection algorithm.
    Loads arbitrarily large 2d image and performs iterative faster rcnn detection on the entire image.

    :param *str* f: path to image by which to analyze
    :param *float* cell_detection_threshold: cells below threshold are rejected
    :param *float* nms_threshold: iou rejection threshold for nms.
    :return: *Cochlea* object containing data of analysis.
    """
    print('Initializing hair cell detection algorithm...')
    if f is None:
        print('\x1b[1;31;40m' + 'ERROR: No File to Analyze... \nAborting.' + '\x1b[0m')
        return None
    if not pixel_size:
        print('\x1b[1;33;40m'
              'WARNING: Pixel Size is not set. Defaults to 288.88 nm x/y. Consider suplying value for optimal performance.'
              '\x1b[0m')

    with torch.no_grad():

        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        if device == 'cuda':
            print('\x1b[1;32;40mCUDA: GPU successfully initialized!\x1b[0m')
        else:
            print('\x1b[1;33;40m'
                  'WARNING: GPU not present or CUDA is not correctly intialized for GPU accelerated computation. '
                  'Analysis may be slow.'
                  '\x1b[0m')

        # Load and preprocess Image
        image_base = load(f, 'TileScan 1 Merged', verbose=True)  # from hcat.lib.utils
        image_base = image_base[[2, 3],...].max(-1) if image_base.ndim == 4 else image_base

        shape = list(image_base.shape)
        shape[0] = 1

        dtype = image_base.dtype if dtype is None else dtype
        scale: int = hcat.lib.utils.get_dtype_offset(dtype)


        temp = np.zeros(shape)
        temp = np.concatenate((temp, image_base)) / scale * 255
        io.imsave(f[:-4:]+'.png', temp.transpose((1,2,0)))

        c, x, y = image_base.shape
        print(
            f'DONE: shape: {image_base.shape}, min: {image_base.min()}, max: {image_base.max()}, dtype: {image_base.dtype}')

        if image_base.max() < scale * 0.33:
            print('\x1b[1;33;40m'
                  f'WARNING: Image max value less than 1/3 the scale factor for bit depth. Image Max: {image_base.max()},'
                  f' Scale Factor: {scale}, dtype: {dtype}. Readjusting scale to 1.5 time Image max.'
                  '\x1b[0m')
            scale = image_base.max() * 1.5



        image_base = torch.from_numpy(image_base.astype(np.uint16) / scale).to(device)

        if pixel_size is not None:
            image_base: Tensor = correct_pixel_size(image_base, pixel_size) #model expects pixel size of 288.88
            print(f'Rescaled Image to match pixel size of 288.88nm with a new shape of: {image_base.shape}')

        elif cell_diameter is not None:
            image_base: Tensor = scale_to_hair_cell_diameter(image_base, cell_diameter)
            print(f'Rescaled Image to match pixel size of 288.88nm with a new shape of: {image_base.shape}')

        # normalize around zero
        image_base.sub_(0.5).div_(0.5)



        # Initalize the model...
        model = FasterRCNN_from_url(url='https://github.com/buswinka/hcat/blob/master/modelfiles/detection.trch?raw=true', device=device)
        model.eval()


        # Initalize curvature detection
        predict_curvature = hcat.lib.functional.PredictCurvature(erode=3)

        # Get the indicies for evaluating cropped regions
        c, x, y = image_base.shape
        image_base = torch.cat((torch.zeros((1, x, y), device=device), image_base), dim=0)
        x_ind: List[List[int]] = calculate_indexes(10, 235, x, x)  # [[0, 255], [30, 285], ...]
        y_ind: List[List[int]] = calculate_indexes(10, 235, y, y)  # [[0, 255], [30, 285], ...]
        total: int = len(x_ind) * len(y_ind)

        # Initalize other small things
        cell_id = 1
        cells = []
        add_cell = cells.append  # stupid but done for speed

        for x, y in tqdm(product(x_ind, y_ind), total=total, desc='Detecting: '):

            # Load and prepare image crop for ML model evaluation
            image: Tensor = image_base[:, x[0]:x[1], y[0]:y[1]].unsqueeze(0)

            # If the image has nothing in it we can skip for speed
            if image.max() == -1:
                continue

            # Evaluate Deep Learning Model
            out: Dict[str, Tensor] = model(image.float())[0]

            scores: Tensor = out['scores'].cpu()
            boxes: Tensor = out['boxes'].cpu()
            labels: Tensor = out['labels'].cpu()

            # The model output coords with respect to the crop of image_base. We have to adjust
            # idk why the y and x are flipped. Breaks otherwise.
            boxes[:, [0, 2]] += y[0]
            boxes[:, [1, 3]] += x[0]

            # center x, center y, width, height
            centers: Tensor = torchvision.ops.box_convert(boxes, 'xyxy', 'cxcywh').cpu()
            cx = centers[:, 0]
            cy = centers[:, 1]

            for i, score in enumerate(scores):
                if score > cell_detection_threshold:
                    add_cell(Cell(id=cell_id,
                                  loc=torch.tensor([0, cx[i], cy[i], 0]),
                                  image=None,
                                  mask=None,
                                  cell_type='OHC' if labels[i] == 1 else 'IHC',
                                  boxes=boxes[i, :],
                                  scores=scores[i]))
                cell_id += 1

        # some cells may overlap. We remove cells after analysis is complete.
        cells: List[Cell] = _cell_nms(cells, nms_threshold)

        ohc = sum([int(c.type == 'OHC') for c in cells])  # number of ohc
        ihc = sum([int(c.type == 'IHC') for c in cells])  # number of ihc
        print(f'Total Cells: {len(cells)}\n   OHC: {ohc}\n   IHC: {ihc}' )

        max_projection: Tensor = image_base[[1], ...].mul(0.5).add(0.5).unsqueeze(-1).cpu()
        curvature, distance, apex = predict_curvature(max_projection, cells, curve_path)

        if curvature is None:
            print('\x1b[1;33;40mWARNING: ' +
                  'All three methods to predict hair cell path have failed. Frequency Mapping functionality is limited.'
                  'Consider Manual Calculation.'
                  + '\x1b[0m')

        # curvature estimation really only works if there is a lot of tissue...
        if distance is not None and distance.max() > 4000:
            for c in cells: c.calculate_frequency(curvature[[0, 1], :], distance)  # calculate cell's best frequency
            cells = [c for c in cells if not c._distance_is_far_away]  # remove a cell if its far away from curve

        else:
            curvature, distance, apex = None, None, None
            print('\x1b[1;33;40mWARNING: ' +
                  'Predicted Cochlear Distance is below 4000um. Not sufficient information to determine cell frequency.'
                  + '\x1b[0m')

        xml = get_xml(f) if f.endswith('.lif') else None
        filename = os.path.split(f)[-1]

        # Store in compressible object for further use
        c = Cochlea(mask=None,
                    filename=filename,
                    path=f,
                    analysis_type='detect',
                    leica_metadata=xml,
                    im_shape=image_base.shape,
                    cochlear_distance=distance,
                    curvature=curvature,
                    cells=cells,
                    apex=apex)

        if save_xml: cochlea_to_xml(c)

        if save_fig: c.make_detect_fig(image_base)
        # c.make_cochleogram()
        print('')
        return c


def _cell_nms(cells: List[Cell], nms_threshold: float) -> List[Cell]:
    """
    Perforns non maximum supression on the resulting cell predictions

    :param cells: Iterable of cells
    :param nms_threshold: cell iou threshold
    :return: Iterable of cells
    """
    # nms to get rid of cells
    boxes = torch.zeros((len(cells), 4))
    scores = torch.zeros(len(cells))
    for i, c in enumerate(cells):
        boxes[i, :] = c.boxes
        scores[i] = c.scores

    ind = torchvision.ops.nms(boxes, scores, nms_threshold)

    # need to pop off list elements from an int64 tensor
    ind_bool = torch.zeros(len(cells))
    ind_bool[ind] = 1
    for i, val in enumerate(ind_bool):
        if val == 0:
            cells[i] = None

    return [c for c in cells if c]



import torch.nn as nn
import wget
import os.path
import torch
import hcat
import re

class Backend(nn.Module):
    def __init__(self):
        super(Backend, self).__init__()

        self.image_reject=True

    def no_reject(self):
        """
        Disables image rejection; forces backend model to FULLY evaluate every image.

        Image rejection is on by default, disabling may significantly degrade overall performance.

        Example:

        >>> from hcat.backends.spatial_embedding import SpatialEmbedding
        >>> backend = SpatialEmbedding()
        >>>
        >>> url = 'https://www.model_location.com/model.trch'
        >>> backend.load(url) # Works with url
        >>>
        >>> model_path = 'path/to/my/model.trch'
        >>> backend.load(model_path) # Also works with path


        """
        self.image_reject = False

    def reject(self):
        """
        Enables image rejection criteria. May result in significant whole cochlea speedup.

        Image rejection is on by default, significantly improving whole image analysis speed.
        """
        self.image_reject = True

    @staticmethod
    @torch.jit.script
    def _is_image_bad(image: torch.Tensor, min_threshold: float = 0.05):
        """
        Check if an image is likely to NOT contain any cells.
        Uses cytosolic stain threshold.

        :param image: input torch tensor
        :param min_threshold: minimum value as percentage of saturated voxels
        :return:
        """
        is_bad = False

        brightness_threshold = torch.tensor(3500).div(2 ** 16).sub(0.5).div(0.5)

        if image.max() == -1:
            is_bad = True
        elif torch.sum(image.gt(brightness_threshold)) < (image.numel() * min_threshold):
            is_bad = True

        return is_bad

    @staticmethod
    def _model_loader_path(path: str, model, device: str):
        """ Loads model from a path """
        try:
            model = torch.jit.script(model).to(device)
            if path is not None:
                checkpoint = torch.load(path)
                if isinstance(checkpoint, dict):
                    checkpoint = checkpoint['model_state_dict']
                model.load_state_dict(checkpoint)
                print('src.lib.backend._model_loader_path: model successfully loaded.')

        except RuntimeError:  # This is likely due to model weights not lining up.
            model = torch.jit.script(model(1, 4).requires_grad_(False)).to(device)
            if path is not None:
                checkpoint = torch.load(path)
                if isinstance(checkpoint, dict):
                    checkpoint = checkpoint['model_state_dict']
                model.load_state_dict(checkpoint)

        for m in model.modules():
            if isinstance(m, nn.BatchNorm3d):
                m.eval()

        return model

    @staticmethod
    def _model_loader_url(url: str, model, device: str):
        """ loads model from url """
        path = os.path.join(hcat.__path__[0], 'spatial_embedding.trch')


        if not os.path.exists(path):
            print('Downloading Model File: ')
            wget.download(url=url, out=path)
            print(' ')

        model = torch.jit.script(model(in_channels=1).requires_grad_(False))
        checkpoint = torch.load(path, map_location=torch.device('cpu'))
        model = model.to(device)
        model.load_state_dict(checkpoint['model_state_dict'])

        for m in model.modules():
            if isinstance(m, nn.BatchNorm3d):
                m.eval()

        return model

    def _is_url(self, input: str):
        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)

        # Return true if its a url
        return re.match(regex, input) is not None


    @staticmethod
    def _colormask_to_mask(mask: torch.Tensor) -> torch.Tensor:
        """
        Converts a integer mask from the watershed algorithm to a 4d matrix where tensor.shape[1] is the number of
        unique cells in the mask.

        :param mask: [B, 1, X, Y, Z]  where each integer is a unique cell
        :return: [B, N, X, Y, Z] where each N is a single cell
        """
        b, _, x, y, z = mask.shape
        n = len(mask.unique()) - 1  # subtract 1 because background is included
        n = n if n > 0 else 0

        out = torch.zeros((b, n, x, y, z))
        unique = mask.unique()
        unique = unique[unique != 0]

        for i, u in enumerate(unique):
            if u == 0:
                continue
            out[0, i, ...] = (mask[0, 0, ...] == u).float()

        return out




import torch
import torch.nn as nn
from warnings import filterwarnings

filterwarnings("ignore", category=UserWarning)


class HCBlock(nn.Module):
    def __init__(self, in_channels: int, kernel_size: int = 5, padding: int = 2) -> None:
        super(HCBlock, self).__init__()

        self.conv1x1 = nn.Conv3d(in_channels * 2, in_channels, kernel_size=1)
        self.conv3x3 = nn.Conv3d(in_channels, in_channels, kernel_size=kernel_size, padding=padding, stride=1, dilation=1)
        self.activation = nn.LeakyReLU()

        self.batch_norm_conv = nn.BatchNorm3d(in_channels)
        self.batch_norm_group = nn.BatchNorm3d(in_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        x = self.activation(self.batch_norm_conv(self.conv1x1(x)))
        x = self.activation(self.batch_norm_group(self.conv3x3(x)))

        return x



import torch
import torchvision.transforms.functional
import torch.nn.functional as F
from kornia.augmentation import RandomAffine3D
from PIL.Image import Image
import numpy as np
from typing import Dict, Tuple, Union, Sequence, List
import elasticdeform
import skimage.io as io
from hcat import ShapeError
import matplotlib.pyplot as plt


# -------------------------------- Assumptions ------------------------------#
#                Every image is expected to be [C, X, Y, Z]                  #
#        Every transform's input has to be Dict[str, torch.Tensor]           #
#       Every transform's output has to be Dict[str, torch.Tensor]           #
#       Preserve whatever device the tensor was on when it started           #
# ---------------------------------------------------------------------------#

class transform:
    """ parent class to every transform. """
    def __init__(self):
        pass
    def check_inputs(self, x: Dict[str, torch.Tensor]) -> None:
        """
        Checks inputs of a transform and raises errors if they are bad.

        Necessary dict keys:
            'image': [C, X, Y, Z] torch.Tensor
            'masks': [C, X, Y, Z] torch.Tensor with identical x,y,z to image
            'centroids': [C, K=3] torch.Tensor with C number of objects

        :param x: Dict[str, torch.Tensor] input to transform
        :return: None
        :raises: RuntimeError: Raises Error if shapes are not expected
        :raises: ValueError: Raises Error if key is not a string, or value is not a tensor
        :raises: KeyError: Raises Error if expected keys are not in the dictionary
        """
        if not isinstance(x, dict): raise ValueError(f'Expected input to be Dict not {type(x)}')

        for key in x:
            if not isinstance(key, str):
                raise ValueError(f'Key: {key} in input dictionary is not type str, found type: {type(key)}')

        for key in x:
            if not isinstance(x[key], torch.Tensor):
                raise ValueError(f'Value for key: {key} in input dictionary is not of type torch.Tensor, found type: {type(x[key])}')

        if 'image' not in x: raise KeyError('key "image" not found in input dictionary.')
        if 'masks' not in x: raise KeyError('key "masks" not found in input dictionary.')
        if 'centroids' not in x: raise KeyError('key "centroids" not found in input dictionary.')

        if x['image'].shape[1::] != x['masks'].shape[1::] and x['masks'].numel() != 0:
            raise ShapeError(f'Shape of input["image"]: {x["image"].shape} != '
                               f'shape of input["masks"]: {x["masks"].shape}')

        for key in x:
            if x[key].device != x['image'].device:
                raise RuntimeError(f'Tensors of key: "{key}" do not share same device.\n'
                                   f'\t"image": {x["image"].device}\n'
                                   f'\t"masks": {x["masks"].device}\n'
                                   f'\t"centroids": {x["centroids"].device}\n')

        if x['image'].ndim != 4: raise ShapeError(f'Expected image.ndim = 4, not: {x["image"].ndim}')
        if x['masks'].ndim != 4: raise ShapeError(f'Expected mask.ndim = 4, not: {x["masks"].ndim} ')


class adjust_brightness(transform):
    def __init__(self, rate: float = 0.5, range_brightness: Tuple[float, float] = (-0.5, 0.5)) -> None:
        super(adjust_brightness, self).__init__()
        self.rate = rate
        self.range = range_brightness
        self.fun = self._adjust_brightness

    def __call__(self, data_dict):
        self.check_inputs(data_dict)

        if torch.rand(1) < self.rate:
            # funky looking but FAST
            val = torch.FloatTensor(data_dict['image'].shape[0]).uniform_(self.range[0], self.range[1])
            data_dict['image'] = self.fun(data_dict['image'], val)

        return data_dict

    @staticmethod
    @torch.jit.script
    def _adjust_brightness(img: torch.Tensor, val: torch.Tensor) -> torch.Tensor:
        """
        Adjusts brigtness of img with val

        :param img: [C, X, Y, Z]
        :param val: Tensor[float] [C]
        :return:
        """
        img = img.add_(val.reshape(img.shape[0], 1, 1, 1).to(img.device))
        img[img < 0] = 0
        img[img > 1] = 1
        return img


class adjust_gamma(transform):
    def __init__(self, rate: float = 0.5, gamma: Tuple[float, float] = (0.5, 1.5),
                 gain: Tuple[float, float] = (.75, 1.25)) -> None:
        super(adjust_gamma, self).__init__()
        self.rate = rate
        self.gain = gain
        self.gamma = gamma

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Randomly adjusts gamma of image color channels independently

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)

        if torch.rand(1) < self.rate:
            gamma = torch.FloatTensor(data_dict['image'].shape[0]).uniform_(self.gamma[0], self.gamma[1])
            gain = torch.FloatTensor(data_dict['image'].shape[0]).uniform_(self.gain[0], self.gain[1])

            data_dict['image'] = self._adjust_gamma(data_dict['image'], gamma, gain)

        return data_dict

    @staticmethod
    @torch.jit.script
    def _adjust_gamma(img: torch.Tensor, gamma: torch.Tensor, gain: torch.Tensor) -> torch.Tensor:
        """
        Assume img in shape [C, X, Y, Z]

        :param img:
        :param gamma:
        :param gain:
        :return:
        """
        for c in range(img.shape[0]):
            img[c, ...] = torchvision.transforms.functional.adjust_gamma(img[c, ...], gamma=gamma[c], gain=gain[c])
        return img


class affine3d(transform):
    def __init__(self,
                 rate: float = 0.75,
                 angle: Dict[str, Tuple[float, float]] = {'yaw': (-180., 180.), 'pitch': (-20., 20.), 'roll': (-20., 20.)},
                 shear: Tuple[float, float] = (-5., 5.),
                 scale: Tuple[float, float] = (0.8, 1.2)) -> None:
        """
        yaw: rotate around Y
        pitch: rotate around x
        roll: rotate around z

        :param rate:
        :param angle:
        :param shear:
        :param scale:
        """
        super(affine3d, self).__init__()
        self.rate = rate
        self.angle = angle
        self.shear = shear
        self.scale = scale

        for key in self.angle:
            if not isinstance(self.angle[key][0], float):
                raise ValueError(f'Angle must be a float: {key}, {self.angle[key][0]}')
            if not isinstance(self.angle[key][1], float):
                raise ValueError(f'Angle must be a float: {key}, {self.angle[key][1]}')

        s = (self.angle['yaw'], self.angle['pitch'], self.angle['roll'])
        self.fun = RandomAffine3D(degrees=s, shears=self.shear, scale=self.scale, resample=0, p=1)

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Performs an affine transformation on the image and mask
        Shears, rotates and scales the image randomly based on parameters defined at initialization

        Adjusts for anisotropy stupidly....
        performs a 2d version if data_dict['image'].shape[-1] == 1

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)
        if torch.rand(1) < self.rate:
            if data_dict['image'].shape[-1] > 1:
                stack = torch.cat((data_dict['image'], data_dict['masks']), dim=0).unsqueeze(0)
                b, c, x, y, z = stack.shape
                bigstack = torch.zeros((b, c, x, y, z * 3))

                # stacks are nonisotropic. Hacky way to adjust for this.
                for i in range(z):
                    bigstack[..., (3 * i) + 0] = stack[..., i]
                    bigstack[..., (3 * i) + 1] = stack[..., i]
                    bigstack[..., (3 * i) + 2] = stack[..., i]

                bigstack = self.fun(bigstack.squeeze(0))                # Perform affine on expanded stack

                for i in range(z):
                    stack[..., i] = bigstack[..., (3 * i) + 1]          # Return to nonisotropic

                data_dict['image'] = stack[0, 0:-1:1, ...]
                data_dict['masks'] = stack[0, [-1], ...]

            else:
                # 2D implementation
                angle = torch.FloatTensor(1).uniform_(self.angle['yaw'][0], self.angle['yaw'][1])
                shear = torch.FloatTensor(1).uniform_(self.shear[0], self.shear[1])
                scale = torch.FloatTensor(1).uniform_(self.scale[0], self.scale[1])
                data_dict['image'] = _reshape(self._affine2d(_shape(data_dict['image']), angle, scale, shear))
                data_dict['masks'] = _reshape(self._affine2d(_shape(data_dict['masks']), angle, scale, shear))

        return data_dict

    @staticmethod
    @torch.jit.script
    def _affine2d(img: torch.Tensor, angle: torch.Tensor,
                  scale: torch.Tensor, shear: torch.Tensor) -> torch.Tensor:
        """
        Assume img in shape [C, Z, X, Y] # FROM _shape

        :param img:
        :param gamma:
        :param gain:
        :return:
        """
        for c in range(img.shape[0]):
            img[c, ...] = torchvision.transforms.functional.affine(img[c, ...], angle=angle.item(), shear=[float(shear.item())],
                                                                   scale=scale.item(), translate=[0,0])
        return img


class colormask_to_mask(transform):
    def __init__(self):
        super(colormask_to_mask, self).__init__()

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Some transformations are done on "colormasks" in which the mask is a integer tensor with values between 1 and N
        objects in the image. This function transforms the "colormask" to a boolean tensor of shape [N,X,Y,Z] where N is the
        number of objects. Pixels assigned a value of True belong to that object.

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)
        data_dict['masks'] = self._colormask_to_torch_mask(data_dict['masks'])
        return data_dict

    @staticmethod
    @torch.jit.script
    def _colormask_to_torch_mask(colormask: torch.Tensor) -> torch.Tensor:
        """

        :param colormask: [C=1, X, Y, Z]
        :return:
        """
        uni = torch.unique(colormask)
        uni = uni[uni != 0]
        num_cells = len(uni)

        shape = (num_cells, colormask.shape[1], colormask.shape[2], colormask.shape[3])
        mask = torch.zeros(shape, device=colormask.device)

        for i, u in enumerate(uni):
            mask[i, :, :, :] = (colormask[0, :, :, :] == u)

        return mask


class debug:
    def __init__(self, ind: int = 0):
        self.ind = ind

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        image = data_dict['image']
        mask = data_dict['masks']
        try:
            assert image.shape[-1] == mask.shape[-1]
            assert image.shape[-2] == mask.shape[-2]
            assert image.shape[-3] == mask.shape[-3]

            assert image.max() <= 1
            assert mask.max() <= 1
            assert image.min() >= 0
            assert mask.min() >= 0
        except Exception as ex:
            print(self.ind)
            raise ex

        return data_dict


class elastic_deformation(transform):
    def __init__(self, grid_shape: Tuple[int, int, int] = (2, 2, 2), scale: int = 2, rate: float = 0.5):
        super(elastic_deformation, self).__init__()
        self.x_grid = grid_shape[0]
        self.y_grid = grid_shape[1]
        self.z_grid = grid_shape[2] if len(grid_shape) > 2 else None
        self.scale = scale
        self.rate = rate

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        self.check_inputs(data_dict)

        device = data_dict['image'].device
        image = data_dict['image'].cpu().numpy()
        mask = data_dict['masks'].cpu().numpy()
        dtype = image.dtype

        if torch.rand(1) < self.rate:

            # 3D
            if image.shape[-1] != 1:
                displacement = np.random.randn(3, self.x_grid, self.y_grid, self.z_grid) * self.scale
                image = elasticdeform.deform_grid(image, displacement, axis=(1, 2, 3))
                mask = elasticdeform.deform_grid(mask, displacement, axis=(1, 2, 3), order=0)

                image[image < 0] = 0.0
                image[image > 1] = 1.0
                image.astype(dtype)
                data_dict['image'] = torch.from_numpy(image).to(device)
                data_dict['masks'] = torch.from_numpy(mask).to(device)

            # 2D
            elif image.shape[-1] == 1:
                displacement = np.random.randn(2, self.x_grid, self.y_grid) * self.scale
                image = elasticdeform.deform_grid(image[..., 0], displacement, axis=(1, 2))
                mask = elasticdeform.deform_grid(mask[..., 0], displacement, axis=(1, 2), order=0)

                image[image < 0] = 0.0
                image[image > 1] = 1.0
                image.astype(dtype)
                data_dict['image'] = torch.from_numpy(image).to(device).unsqueeze(-1) # correct for z reduction
                data_dict['masks'] = torch.from_numpy(mask).to(device).unsqueeze(-1)



        return data_dict


class erosion(transform):
    def __init__(self, kernel_targets: int = 3, rate: float = 0.5, device: str = 'cpu') -> None:
        super(erosion, self).__init__()
        self.device = device
        self.rate = rate
        self.kernel_targets = kernel_targets

        if kernel_targets % 2 != 1:
            raise ValueError('Expected Kernel target to be Odd')

        self.padding: Tuple[int, int, int] = self._compute_zero_padding(
            (kernel_targets, kernel_targets, kernel_targets))
        self.kernel: torch.Tensor = self._get_binary_kernel3d(kernel_targets, device)

    def __call__(self, input: Union[Dict[str, torch.Tensor], torch.Tensor]) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
        """
        Randomly applies a gaussian blur

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictionary with identical keys as input, but with transformed values
        """
        if isinstance(input, Dict):
            self.check_inputs(input)
            return self._is_dict(input)
        elif isinstance(input, torch.Tensor):
            return self._is_tensor(input)

    def _is_dict(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        if data_dict['masks'].dtype != self.kernel.dtype:
            raise ValueError(f'Expected Image dtype to be {self.kernel.dtype} not {data_dict["masks"].dtype}')
        # We expect the image to have ndim==4 but need a batch size of 1 for median filter

        device = data_dict['image'].device
        if self.kernel.device != device:
            self.kernel = self.kernel.to(device)

        if torch.rand(1) < self.rate:
            im = data_dict['masks'].unsqueeze(0)
            b, c, h, w, d = im.shape
            # map the local window to single vector
            features: torch.Tensor = F.conv3d(im.reshape(b * c, 1, h, w, d),
                                              self.kernel,
                                              padding=self.padding, stride=1)
            data_dict['masks'] = features.min(dim=1)[0]
        return data_dict


    def _is_tensor(self, input: torch.Tensor) -> torch.Tensor:
        "Assumes [C, X, Y, Z]"
        if input.device != self.kernel.device:
            raise ValueError(f'Expected Image Device to be {self.kernel.device} not {input.device}')
        if input.dtype != self.kernel.dtype:
            raise ValueError(f'Expected Image dtype to be {self.kernel.dtype} not {input.dtype}')
        if input.ndim != 4:
            raise ValueError(f'Expected Image ndim to be 4 not {input.ndim}, with shape {input.shape}')

        b, c, h, w, d = input.unsqueeze(0).shape
        # map the local window to single vector
        features: torch.Tensor = F.conv3d(input.reshape(b * c, 1, h, w, d), self.kernel,
                                          padding=self.padding, stride=1)
        return features.min(dim=1)[0]

    @staticmethod
    def _compute_zero_padding(kernel_size: Tuple[int, int, int]) -> Tuple[int, int, int]:
        r"""Utility function that computes zero padding tuple.
        Adapted from Kornia
        """
        computed: List[int] = [(k - 1) // 2 for k in kernel_size]
        return computed[0], computed[1], computed[2]

    @staticmethod
    def _get_binary_kernel3d(window_size: int, device: torch.device) -> torch.Tensor:
        r"""Creates a symetric binary kernel to extract the patches. If the window size
        is HxWxD will create a (H*W)xHxW kernel.

        ADAPTED FROM KORNIA

        """
        window_range: int = int(window_size ** 3)
        kernel: torch.Tensor = torch.zeros((window_range, window_range, window_range), device=device)
        for i in range(window_range):
            kernel[i, i, i] += 1.0
        kernel = kernel.view(-1, 1, window_size, window_size, window_size)

        # get rid of all zero kernels
        ind = torch.nonzero(kernel.view(kernel.shape[0], -1).sum(1))
        return kernel[ind[:, 0], ...]


class gaussian_blur(transform):
    def __init__(self, kernel_targets: torch.Tensor = torch.tensor([3, 5, 7]), rate: float = 0.5) -> None:
        super(gaussian_blur, self).__init__()
        self.kernel_targets = kernel_targets
        self.rate = rate
        self.fun = torch.jit.script(torchvision.transforms.functional.gaussian_blur)

    def __call__(self, data_dict):
        """
        Randomly applies a gaussian blur

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictionary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)

        if torch.rand(1) < self.rate:
            kern = self.kernel_targets[int(torch.randint(0, len(self.kernel_targets), (1, 1)).item())].item()
            data_dict['image'] = _reshape(self.fun(_shape(data_dict['image']), [kern, kern]))
        return data_dict


class median_filter(transform):
    def __init__(self, kernel_targets: int = 3, rate: float = 0.5, device: str = 'cpu') -> None:
        super(median_filter, self).__init__()
        self.device = device
        self.rate = rate
        self.kernel_targets = kernel_targets

        if kernel_targets % 2 != 1:
            raise ValueError('Expected Kernel target to be Odd')

        self.padding: Tuple[int, int, int] = self._compute_zero_padding(
            (kernel_targets, kernel_targets, kernel_targets))
        self.kernel: torch.Tensor = self._get_binary_kernel3d(kernel_targets, device)

    def __call__(self, input: Union[Dict[str, torch.Tensor], torch.Tensor]) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
        """
        Randomly applies a gaussian blur

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictionary with identical keys as input, but with transformed values
        """
        if isinstance(input, Dict):
            self.check_inputs(input)
            return self._is_dict(input)
        elif isinstance(input, torch.Tensor):
            return self._is_tensor(input)
        else:
            raise ValueError(f'Transformation not implemented for type: {type(input)}')

    def _is_dict(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        if data_dict['image'].dtype != self.kernel.dtype:
            raise ValueError(f'Expected Image dtype to be {self.kernel.dtype} not {data_dict["image"].dtype}')

        device = data_dict['image'].device
        if self.kernel.device != device:
            self.kernel = self.kernel.to(device)

        if torch.rand(1) < self.rate:
            data_dict['image'] = data_dict['image'].unsqueeze(0)
            b, c, h, w, d = data_dict['image'].shape
            # map the local window to single vector
            features: torch.Tensor = F.conv3d(data_dict['image'].reshape(b * c, 1, h, w, d), self.kernel,
                                              padding=self.padding, stride=1)
            data_dict['image'] = torch.median(features.view(b, c, -1, h, w, d), dim=2)[0].squeeze(0)

        return data_dict

    def _is_tensor(self, input: torch.Tensor) -> torch.Tensor:
        "Assume [B, C, X, Y, Z]"
        if input.device != self.kernel.device:
            raise ValueError(f'Expected Image Device to be {self.kernel.device} not {input.device}')
        if input.dtype != self.kernel.dtype:
            raise ValueError(f'Expected Image dtype to be {self.kernel.dtype} not {input.dtype}')
        if input.ndim != 5:
            raise ValueError(f'Expected Image ndim to be 5 not {input.ndim}, with shape {input.shape}')

        b, c, h, w, d = input.shape
        # map the local window to single vector
        features: torch.Tensor = F.conv3d(input.reshape(b * c, 1, h, w, d), self.kernel,
                                          padding=self.padding, stride=1)
        return torch.median(features.view(b, c, -1, h, w, d), dim=2)[0]

    @staticmethod
    def _compute_zero_padding(kernel_size: Tuple[int, int, int]) -> Tuple[int, int, int]:
        r"""Utility function that computes zero padding tuple.
        Adapted from Kornia
        """
        computed: List[int] = [(k - 1) // 2 for k in kernel_size]
        return computed[0], computed[1], computed[2]

    @staticmethod
    def _get_binary_kernel3d(window_size: int, device: torch.device) -> torch.Tensor:
        r"""Creates a symetric binary kernel to extract the patches. If the window size
        is HxWxD will create a (H*W)xHxW kernel.

        ADAPTED FROM KORNIA

        """
        window_range: int = int(window_size ** 3)
        kernel: torch.Tensor = torch.zeros((window_range, window_range, window_range), device=device)
        for i in range(window_range):
            kernel[i, i, i] += 1.0
        kernel = kernel.view(-1, 1, window_size, window_size, window_size)

        # get rid of all zero kernels
        ind = torch.nonzero(kernel.view(kernel.shape[0], -1).sum(1))
        return kernel[ind[:, 0], ...]


class normalize(transform):
    def __init__(self, mean: Sequence[float] = [0.5], std: Sequence[float] = [0.5]) -> None:
        super(transform)
        self.mean = mean
        self.std = std
        self.fun = torch.jit.script(torchvision.transforms.functional.normalize)

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """


        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)

        data_dict['image'] = self.fun(data_dict['image'], self.mean, self.std)
        return data_dict


class nul_crop(transform):
    def __init__(self, rate: float = 0.80, limit_z: int = 3) -> None:
        super(nul_crop, self).__init__()
        self.rate = rate
        self.limit_z = limit_z

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Removes blank space around cells to ensure training images has something the network can learn
        Doesnt remove Z


        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)

        shape = data_dict['masks'].shape

        if torch.rand(1) < self.rate:
            ind = torch.nonzero(data_dict['masks'])  # -> [I, 4] where 4 is ndims
            x_max = ind[:, 1].max().int().item()
            y_max = ind[:, 2].max().int().item()
            z_max = ind[:, 3].max().int().item()

            x = ind[:, 1].min().int().item()
            y = ind[:, 2].min().int().item()
            z = ind[:, 3].min().int().item() - self.limit_z
            z = z if z > 0 else 0

            w = x_max - x
            h = y_max - y
            d = z_max + self.limit_z - z
            d = d if d < shape[-1] else shape[-1]

            data_dict['image'] = _crop(data_dict['image'], x=x, y=y, z=0, w=w, h=h, d=d)
            data_dict['masks'] = _crop(data_dict['masks'], x=x, y=y, z=0, w=w, h=h, d=d)

        return data_dict


class random_affine(transform):
    def __init__(self, rate: float = 0.5, angle: Tuple[int, int] = (-180, 180),
                 shear: Tuple[int, int] = (-5, 5), scale: Tuple[float, float] = (0.9, 1.1)) -> None:
        super(random_affine, self).__init__()
        self.rate = rate
        self.angle = angle
        self.shear = shear
        self.scale = scale

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Performs an affine transformation on the image and mask
        Shears, rotates and scales the image randomly based on parameters defined at initialization

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)
        if torch.rand(1) < self.rate:
            angle = torch.FloatTensor(1).uniform_(self.angle[0], self.angle[1])
            shear = torch.FloatTensor(1).uniform_(self.shear[0], self.shear[1])
            scale = torch.FloatTensor(1).uniform_(self.scale[0], self.scale[1])
            translate = torch.tensor([0, 0])

            data_dict['image'] = _reshape(self._affine(_shape(data_dict['image']), angle, translate, scale, shear))
            data_dict['masks'] = _reshape(self._affine(_shape(data_dict['masks']), angle, translate, scale, shear))

        return data_dict

    @staticmethod
    @torch.jit.script
    def _affine(img: torch.Tensor, angle: torch.Tensor, translate: torch.Tensor, scale: torch.Tensor,
                shear: torch.Tensor) -> torch.Tensor:
                                                         """
                                                         Not to be publicly accessed! Only called through hcat.transforms.affine

                                                         A jit scriptable wrapped version of torchvision.transforms.functional.affine
                                                         Cannot, by rule, pass a dict to a torchscript function, necessitating this function


                                                         WARNING: Performs an affine transformation on the LAST TWO DIMENSIONS
                                                         ------------------------------------------------------------------------------------------------------------
                                                         call _shape(img) prior to _affine such that this transformation is performed on the X and Y dimensions
                                                         call _reshape on the output of _affine to return to [C, X, Y, Z]

                                                         correct implementation looks like
                                                         ```python
                                                         from hcat.transforms import _shape, _reshape, _affine
                                                         angle = torch.tensor([0])
                                                         scale = torch.tensor([0])
                                                         shear = torch.tensor([0])
                                                         translate = torch.tensor([0])
                                                         transformed_image = _reshape(_affine(_shape(img), angle, translate, scale, shear))

                                                         ```


                                                         :param img: torch.Tensor from data_dict of shape [..., X, Y]
                                                         :param angle: torch.Tensor float in degrees
                                                         :param translate: torch.Tensor translation factor. If zero, any transformations are done around center of image
                                                         :param scale: torch.Tensor float Scale factor of affine transformation, if 1, no scaling is performed
                                                         :param shear: torch.Tensor float shear factor of affine transformation, if 0, no shearing is performed
                                                         :return: torch.Tensor
                                                         """
                                                         angle = float(angle.item())
                                                         scale = float(scale.item())
                                                         shear = [float(shear.item())]
                                                         translate_list = [int(translate[0].item()), int(translate[1].item())]
                                                         return torchvision.transforms.functional.affine(img, angle, translate_list, scale, shear)


class random_crop(transform):
    def __init__(self, shape: Tuple[int, int, int] = (256, 256, 26)) -> None:
        super(random_crop, self).__init__()
        self.w = shape[0]
        self.h = shape[1]
        self.d = shape[2]

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Randomly crops an image to a designated size. If the crop size is too big, it just takes as much as it can.

        example:
            >>> import torch
            >>> from hcat.train.transforms import random_crop
            >>>
            >>> in_image = torch.rand((300, 150, 27))
            >>> masks = torch.rand((300, 150, 27)).gt(0.5)
            >>> transform = random_crop(shape = (256, 256, 26))
            >>> in_data = {'image': in_image, 'mask': masks, 'centroids': torch.Tensor([])}
            >>> out_data: Dict[str, torch.Tensor] = transform(in_data)
            >>> assert out_data['image'].shape =
                (256, 150, 26epoch_range.desc = 'Loss: ' + '{:.5f}'.format(torch.tensor(epoch_loss).mean().item()))


        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values

        :raises: RuntimeError | Trys to find a valid image. Throws error after 10 failed attemts.
        """
        try:
            self.check_inputs(data_dict)
        except:
            print(data_dict['image'].shape)
            raise AttributeError

        shape = data_dict['image'].shape

        x_max = shape[1] - self.w if shape[1] - self.w > 0 else 1
        y_max = shape[2] - self.h if shape[2] - self.h > 0 else 1
        z_max = shape[3] - self.d if shape[3] - self.d > 0 else 1

        x = torch.randint(x_max, (1, 1)).item()
        y = torch.randint(y_max, (1, 1)).item()
        z = torch.randint(z_max, (1, 1)).item()

        assert data_dict['masks'].sum() != 0

        # Check if the crop doesnt contain any positive labels.
        # If it does, try generating new points
        # We want to make sure every training image has something to learn
        num_try = 0
        while _crop(data_dict['masks'], x=x, y=y, z=z, w=self.w, h=self.h, d=self.d).sum() == 0:
            num_try += 1

            x = torch.randint(x_max, (1, 1)).item()
            y = torch.randint(y_max, (1, 1)).item()
            z = torch.randint(z_max, (1, 1)).item()

            if num_try > 10:
                raise ValueError("Exceded maximum try's to find a valid image.")


        data_dict['image'] = _crop(data_dict['image'], x=x, y=y, z=z, w=self.w, h=self.h, d=self.d)
        data_dict['masks'] = _crop(data_dict['masks'], x=x, y=y, z=z, w=self.w, h=self.h, d=self.d)

        mask = data_dict['masks']
        c, x, y, z, = mask.shape
        i = mask.reshape(c, -1).sum(-1) != 0
        data_dict['masks'] = data_dict['masks'][i,...]

        return data_dict


class random_h_flip(transform):
    def __init__(self, rate: float = 0.5) -> None:
        super(random_h_flip, self).__init__()
        self.rate = rate
        self.fun = torch.jit.script(torchvision.transforms.functional.hflip)

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Randomly flips the mask vertically.

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)

        if torch.rand(1) < self.rate:
            data_dict['image'] = _reshape(self.fun(_shape(data_dict['image'])))
            data_dict['masks'] = _reshape(self.fun(_shape(data_dict['masks'])))

        return data_dict


class random_v_flip(transform):
    def __init__(self, rate: float = 0.5) -> None:
        super(random_v_flip, self).__init__()
        self.rate = rate
        self.fun = torch.jit.script(torchvision.transforms.functional.vflip)

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Randomly flips the mask vertically.

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)

        if torch.rand(1) < self.rate:
            data_dict['image'] = _reshape(self.fun(_shape(data_dict['image'])))
            data_dict['masks'] = _reshape(self.fun(_shape(data_dict['masks'])))

        return data_dict


class random_noise(transform):
    def __init__(self, gamma: float = 0.1, rate: float = 0.5):
        super(random_noise, self).__init__()
        self.gamma = gamma
        self.rate = rate

    def __call__(self, data_dict: Dict[str, torch.Tensor], ) -> Dict[str, torch.Tensor]:
        """
        Adds noise to the image. Noise are values between 0 and 0.3

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)

        if torch.rand(1).item() < self.rate:
            device = data_dict['masks'].device
            noise = torch.rand(data_dict['image'].shape).to(device) * torch.tensor([self.gamma]).to(device)
            data_dict['image'] = data_dict['image'] + noise

        return data_dict


class save_image:
    def __init__(self, name: str):
        self.name = name

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        image = data_dict['image'][[2, 1, 0], ...].cpu().transpose(0, -1).numpy()[10, ...]
        _, mask = torch.max(data_dict['masks'].cpu(), 0)

        print(mask.max())

        mask = mask.float().numpy().transpose((2, 0, 1))[10, ...]

        io.imsave(self.name + '_image.png', image)
        io.imsave(self.name + '_mask.png', mask)

        return data_dict


class to_cuda(transform):
    def __init__(self):
        pass

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Move every element in a dict containing torch tensor to cuda.

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)
        for key in data_dict:
            data_dict[key] = data_dict[key].cuda()
        return data_dict


class transformation_correction(transform):
    def __init__(self, min_cell_volume: int = 4000):
        super(transformation_correction, self).__init__()
        self.min_cell_volume = min_cell_volume

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Some Geometric transforms may alter the locations of cells so drastically that the centroid may no longer
        be accurate. This recalculates the centroids based on the current mask.

        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks' : torch.Tensor of size [I, X, Y, Z] where I is the number of identifiable objects in the mask
            'centroids' : torch.Tensor of size [I, 3] where dimension two is the [X, Y, Z] position of the centroid
                          for instance i

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys as input, but with transformed values
        """
        self.check_inputs(data_dict)
        shape = data_dict['masks'].shape
        device = data_dict['masks'].device
        centroid = torch.zeros(shape[0], 3, dtype=torch.float, device=device)
        ind = torch.ones(shape[0], dtype=torch.long, device=device)

        for i in range(shape[0]):  # num of instances

            # Selectively remove mask if it is too small
            data_dict['masks'][i, ...] = self._remove_bad_masks(data_dict['masks'][i, ...], self.min_cell_volume)
            indexes = torch.nonzero(data_dict['masks'][i, ...] > 0).float()

            # If every value in a tensor is 0, assign [-1, -1, -1] as the centroid. Will be removed later
            if indexes.shape[0] == 0:
                centroid[i, :] = torch.tensor([-1, -1, -1], device=device)
                ind[i] = 0

            # If there is a mask, figure out its best location
            else:
                z_max = indexes[..., -1].max()
                z_min = indexes[..., -1].min()
                z = torch.round((z_max - z_min) / 2 + z_min) - 2
                indexes = indexes[indexes[..., -1] == z, :]
                centroid[i, :] = torch.cat((torch.mean(indexes, dim=0)[0:2], z.unsqueeze(0))).float()

            if torch.any(torch.isnan(centroid[i, :])):  # Sometimes shit is NaN, if it is, get rid of it
                centroid[i, :] = torch.tensor([-1, -1, -1], device=device)
                ind[i] == 0

        data_dict['centroids'] = centroid[ind.bool()]
        data_dict['masks'] = data_dict['masks'][ind.bool(), :, :, :]

        assert torch.isnan(data_dict['centroids']).sum() == 0

        return data_dict

    @staticmethod
    @torch.jit.script
    def _remove_bad_masks(image: torch.Tensor, min_cell_volume: int) -> torch.Tensor:
        """
        Input an boolean image volume of shape [X, Y, Z] and determine the number of nonzero pixels.

        Sets the segmentation mask of a cell to ZERO if both conditions are met:
            1 - The cell is touching an edge (All edges of volume are considered)
            2 - The cell is smaller than a lower bound set by arg: min_cell_volume

        :param image: [X, Y, Z]
        :param min_cell_volume: min number of voxels needed to keep a cell mask as valid

        :return: torch.Tensor of shape [X, Y, Z]. Returns image if no cells are touching the edge.
        """
        ind = torch.nonzero(image)  # Determine location of nonzero values -> torch.tensor([:, [X, Y, Z]])

        # Loop over each edge (i) and check if the mask is touching.
        # Implicitly, if a nonzero value is zero it is on an edge.
        for i in range(3):
            remove_bool = torch.any(ind[:, i] == 0) or torch.any(ind[:, i] == image.shape[i] - 1)
            remove_bool = remove_bool if torch.sum(image) < min_cell_volume else False

            # Remove cell if it touches the edge and is small. No need for further computation.
            if remove_bool:
                image = torch.zeros(image.shape)
                break

        return image


# Faster RCNN!
class box_to_mask(transform):
    def __init__(self):
        super(box_to_mask, self).__init__()

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Converts boxes to a mask for geometric transformations later!

        NOTE: data_dict must contain two aditional params: 'boxes' and 'labels' !!!!


        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z=1] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks': None
            'centroids' : None
            'boxes': torch.Tensor of size [N,4] where N is the number of objects in the 2D image!
            'boxes': torch.Tensor of size [K] where K is the class label

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys but notable exceptions
            key: val
            'masks': torch.Tensor of size [N, X, Y, Z=1] where X and Y are the same as the input image. N is the number
                     of objects in the image.
        """
        self.check_inputs(data_dict)

        if 'boxes' not in data_dict: raise ValueError('Key: "boxes" not found in input dictonary.')
        # if isinstance(data_dict['boxes'], torch.Tensor): raise ValueError(f'Value pair for key: "boxes" must be of type: torch.Tensor, not: {type(data_dict["boxes"])}')
        if data_dict['boxes'].ndim != 2: raise ShapeError(f'Expected value pair of key: "boxes" to have ndim=2, not {data_dict["boxes"].ndim}')

        device = data_dict['image'].device
        _, x, y, z = data_dict['image'].shape
        n, _ = data_dict['boxes'].shape

        mask = torch.zeros((n, x, y, z), device=device, dtype=torch.int)
        for i in range(n):
            x0, y0, x1, y1 = data_dict['boxes'][i, :]
            mask[i, y0:y1, x0:x1, :] = data_dict['labels'][i] # class label
        data_dict['masks'] = mask

        return data_dict


# Faster RCNN!
class mask_to_box(transform):
    def __init__(self):
        super(mask_to_box, self).__init__()

    def __call__(self, data_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Performs inverse operation of box_to_mask!

        NOTE: data_dict must contain two aditional params: 'boxes' and 'labels' !!!!


        :param data_dict Dict[str, torch.Tensor]: data_dictionary from a dataloader. Has keys:
            key : val
            'image' : torch.Tensor of size [C, X, Y, Z=1] where C is the number of colors, X,Y,Z are the mask height,
                      width, and depth
            'masks': None
            'centroids' : None
            'boxes': torch.Tensor of size [N,4] where N is the number of objects in the 2D image!
            'boxes': torch.Tensor of size [K] where K is the class label

        :return: data_dict Dict[str, torch.Tensor]: dictonary with identical keys but notable exceptions
            key: val
            'masks': torch.Tensor of size [N, X, Y, Z=1] where X and Y are the same as the input image. N is the number
                     of objects in the image.
        """
        self.check_inputs(data_dict)

        if 'boxes' not in data_dict: raise ValueError('Key: "boxes" not found in input dictonary.')
        if data_dict['masks'] is None: raise ValueError('Value pair to key: "masks" is, and must not be, None.')
        # if isinstance(data_dict['boxes'], torch.Tensor): raise ValueError('Value pair for key: "boxes" must be of type: torch.Tensor')
        if data_dict['boxes'].ndim != 2: raise ShapeError(
            f'Expected value pair of key: "boxes" to have ndim=2, not {data_dict["boxes"].ndim}')

        device = data_dict['image'].device
        _, x, y, z = data_dict['image'].shape
        n, _, _, _ = data_dict['masks'].shape

        boxes = torch.zeros([n, 4], device=device)
        labels = torch.zeros([n], device=device)
        for i in range(n):
            nonzero = torch.nonzero(data_dict['masks'][i, ...]) # Q, 3=[x,y,z]
            labels[i] = data_dict['masks'][i,...].max()
            if nonzero.numel() != 0:
                x0 = torch.min(nonzero[:, 1])
                x1 = torch.max(nonzero[:, 1])
                y0 = torch.min(nonzero[:, 0])
                y1 = torch.max(nonzero[:, 0])

                boxes[i, 0] = x0
                boxes[i, 1] = y0
                boxes[i, 2] = x1
                boxes[i, 3] = y1

        data_dict['boxes'] = boxes
        data_dict['masks'] = torch.empty((0,0,0,0), device=device)
        data_dict['labels'] = labels.type(torch.int64)

        return data_dict


# Generic fast helper functions

@torch.jit.script
def _shape(img: torch.Tensor) -> torch.Tensor:
    """
    Shapes a 4D input tensor from shape [C, X, Y, Z] to [C, Z, X, Y]

    * some torchvision functional transforms only work on last two dimensions *

    :param img: torch.Tensor image of shape [C, X, Y, Z]
    :return:
    """
    # [C, X, Y, Z] -> [C, 1, X, Y, Z] ->  [C, Z, X, Y, 1] -> [C, Z, X, Y]
    return img.unsqueeze(1).transpose(1, -1).squeeze(-1)


@torch.jit.script
def _reshape(img: torch.Tensor) -> torch.Tensor:
    """
    Reshapes a 4D input tensor from shape [C, Z, X, Y] to [C, X, Y, Z]

    Performs corrective version of _shape

    * some torchvision functional transforms only work on last two dimensions *

    :param img: torch.Tensor image of shape [C, Z, X, Y]
    :return:
    """
    # [C, Z, X, Y] -> [C, Z, X, Y, 1] ->  [C, 1, X, Y, Z] -> [C, Z, X, Y]
    return img.unsqueeze(-1).transpose(1, -1).squeeze(1)


@torch.jit.script
def _crop(img: torch.Tensor, x: int, y: int, z: int, w: int, h: int, d: int) -> torch.Tensor:
    """
    torch scriptable function which crops an image

    :param img: torch.Tensor image of shape [C, X, Y, Z]
    :param x: x coord of crop box
    :param y: y coord of crop box
    :param z: z coord of crop box
    :param w: width of crop box
    :param h: height of crop box
    :param d: depth of crop box
    :return:
    """
    return img[..., x:x + w, y:y + h, z:z + d]


if __name__ == '__main__':
    pass



import torch
from typing import Dict, Optional, List

# DOCUMENTED

class Sigma:
    """ an adjustable sigma parameter for training. """
    def __init__(self, adjustments: List[Dict[str, float]], initial_sigma = [0.1, 0.1, 0.8], device='cpu'):
        """
        Creates an object which inputs an epoch, and returns a torch.tensor of [sigma_x, sigma_y, sigma_z]

        :param adjustments: lost of adjustments and when to apply them
        :param initial_sigma: initial values of sigma at epoch=0
        :param device: device to load sigma on ('cpu' or 'cuda')
        """
        self.adjutments = adjustments
        self.device = device
        if isinstance(initial_sigma, list):
            self.initial_sigma = torch.tensor(initial_sigma, device=device)
        else:
            self.initial_sigma = initial_sigma

        values = [1]  # Initial values set so sigma isnt zero at 3 zero
        epochs = [-1]
        for d in adjustments:
            values.append(d['multiplier'])
            epochs.append(d['epoch'])

        self.values = torch.tensor(values, device=self.device)
        self.epochs = torch.tensor(epochs, device=self.device)

    def __call__(self, e: int):
        """
        Returns sigma at epoch "e"
        :param e: epoch
        :return: sigma [x,y,z]
        """
        multiplier = self.values[self.epochs < e].prod()

        return self.initial_sigma * multiplier



import torch
from typing import List, Tuple, Optional, Union

import torchvision.transforms.functional

from hcat.lib.explore_lif import Reader
from hcat.train.transforms import _crop

import numpy as np
import skimage.io as io
import os.path

import matplotlib.pyplot as plt
import xml.etree.ElementTree as ET


def graceful_exit(message):
    """ Decorator which returns a message upon failure"""
    def decorator(function):
        def wrapper(*args, **kwargs):
            result = None
            try:
                result = function(*args, **kwargs)
            except Exception:
                print(message)
            return result
        return wrapper
    return decorator


def calculate_indexes(pad_size: int, eval_image_size: int,
                      image_shape: int, padded_image_shape: int) -> List[List[int]]:
    """
    This calculates indexes for the complete evaluation of an arbitrarily large image by unet.
    each index is offset by eval_image_size, but has a width of eval_image_size + pad_size * 2.
    Unet needs padding on each side of the evaluation to ensure only full convolutions are used
    in generation of the final mask. If the algorithm cannot evenly create indexes for
    padded_image_shape, an additional index is added at the end of equal size.

    :param pad_size: int corresponding to the amount of padding on each side of the
                     padded image
    :param eval_image_size: int corresponding to the shape of the image to be used for
                            the final mask
    :param image_shape: int Shape of image before padding is applied
    :param padded_image_shape: int Shape of image after padding is applied

    :return: List of lists corresponding to the indexes
    """

    # We want to account for when the eval image size is super big, just return index for the whole image.
    if eval_image_size + (2 * pad_size) > image_shape:
        return [[0, image_shape - 1]]

    try:
        ind_list = torch.arange(0, image_shape, eval_image_size)
    except RuntimeError:
        raise RuntimeError(f'Calculate_indexes has incorrect values {pad_size} | {image_shape} | {eval_image_size}:\n'
                           f'You are likely trying to have a chunk smaller than the set evaluation image size. '
                           'Please decrease number of chunks.')
    ind = []
    for i, z in enumerate(ind_list):
        if i == 0:
            continue
        z1 = int(ind_list[i - 1])
        z2 = int(z - 1) + (2 * pad_size)
        if z2 < padded_image_shape:
            ind.append([z1, z2])
        else:
            break
    if not ind:  # Sometimes z is so small the first part doesnt work. Check if z_ind is empty, if it is do this!!!
        z1 = 0
        z2 = eval_image_size + pad_size * 2
        ind.append([z1, z2])
        ind.append([padded_image_shape - (eval_image_size + pad_size * 2), padded_image_shape])
    else:  # we always add at the end to ensure that the whole thing is covered.
        z1 = padded_image_shape - (eval_image_size + pad_size * 2)
        z2 = padded_image_shape - 1
        ind.append([z1, z2])
    return ind


@torch.jit.script
def crop_to_identical_size(a: torch.Tensor, b: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Crops Tensor a to the shape of Tensor b, then crops Tensor b to the shape of Tensor a.

    :param a: torch.
    :param b:
    :return:
    """
    # if a.ndim != b.ndim:
    #     raise RuntimeError('Number of dimensions of tensor "a" does not equal tensor "b".')

    if a.ndim < 3:
        raise RuntimeError('Only supports tensors with minimum 3dimmensions and shape [..., X, Y, Z]')

    a = _crop(a, x=0, y=0, z=0, w=b.shape[-3], h=b.shape[-2], d=b.shape[-1])
    b = _crop(b, x=0, y=0, z=0, w=a.shape[-3], h=a.shape[-2], d=a.shape[-1])
    return a, b

# ########################################################################################################################
# #                                                       Postprocessing
# ########################################################################################################################

# @graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not remove edge cells.' + '\x1b[0m')
@torch.jit.script
def remove_edge_cells(mask: torch.Tensor) -> torch.Tensor:
    """
    Removes cells touching the border

    .. warning:
       Will not raise an error upon failure, instead returns None and prints to standard out

    :param mask: (B, X, Y, Z)
    :return: mask (B, X, Y, Z)
    """

    if mask.ndim != 4: raise RuntimeError('input.ndim != 4')

    # Mask is empty, nothing to do.
    if mask.max() == 0:
        return mask

    left = torch.unique(mask[:, 0, :, :])
    right = torch.unique(mask[:, -1, :, :])
    top = torch.unique(mask[:, :, 0, :])
    bottom = torch.unique(mask[:, :, -1, :])

    cells = torch.unique(torch.cat((left, right, top, bottom)))

    for c in cells:
        if c == 0:
            continue
        mask[mask == c] = 0

    return mask


# @graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not remove improperly sized cells.' + '\x1b[0m')
@torch.jit.script
def remove_wrong_sized_cells(mask: torch.Tensor) -> torch.Tensor:
    """
    Removes cells with outlandish volumes. These have to be wrong.

    .. warning:
       Will not raise an error upon failure, instead returns None and prints to standard out

    :param mask: [B, C=1, X, Y, Z] int torch.Tensor: cell segmentation mask where each cell has a unique cell id
    :return:
    """
    unique = torch.unique(mask)
    unique = unique[unique.nonzero()]

    for u in unique:
        if (mask == u).sum() < 4000:
            mask[mask == u] = 0
        elif (mask == u).sum() > 30000:
            mask[mask == u] = 0
    return mask


# ########################################################################################################################
# #                                                       U Net Specific
# ########################################################################################################################


def pad_image_with_reflections(image: torch.Tensor, pad_size: Tuple[int] = (30, 30, 6)) -> torch.Tensor:
    """
    Pads image according to Unet spec
    expect [B, C, X, Y, Z]
    Adds pad size to each side of each dim. For example, if pad size is 10, then 10 px will be added on top, and on bottom.

    :param image:
    :param pad_size:
    :return:
    """
    if not isinstance(image, torch.Tensor):
        raise TypeError(f'Expected image to be of type torch.tensor not {type(image)}')
    for pad in pad_size:
        if pad % 2 != 0:
            raise ValueError('Padding must be divisible by 2')

    device = image.device

    image_size = image.shape
    pad_size = np.array(pad_size)

    left_pad = image.cpu().numpy()[:, :, pad_size[0] - 1::-1, :, :]
    left_pad = torch.as_tensor(left_pad.copy())
    right_pad = image.cpu().numpy()[:, :, -1:-pad_size[0] - 1:-1, :, :]
    right_pad = torch.as_tensor(right_pad.copy())
    image = torch.cat((left_pad, image.cpu(), right_pad), dim=2).to(device)

    left_pad = 0
    right_pad = 0

    bottom_pad = image.cpu().numpy()[:, :, :, pad_size[1] - 1::-1, :]
    bottom_pad = torch.as_tensor(bottom_pad.copy())
    top_pad = image.cpu().numpy()[:, :, :, -1:-pad_size[1] - 1:-1, :]
    top_pad = torch.as_tensor(top_pad.copy())
    image = torch.cat((bottom_pad, image.cpu(), top_pad), dim=3).to(device)
    bottom_pad = 0
    top_pad = 0

    bottom_pad = image.cpu().numpy()[:, :, :, :, pad_size[2] - 1::-1]
    bottom_pad = torch.as_tensor(bottom_pad.copy())
    top_pad = image.cpu().numpy()[:, :, :, :, -1:-pad_size[2] - 1:-1]
    top_pad = torch.as_tensor(top_pad.copy())

    return torch.cat((bottom_pad, image.cpu(), top_pad), dim=4).to(device)


########################################################################################################################
#                                                         Generics
########################################################################################################################


def load(file: str,  header_name: Optional[str] = 'TileScan 1 Merged',
         verbose: bool = False) -> Union[None, np.array]:
    """
    Loads image file (*leica or *tif) and returns an np.array

    :param file: str path to the file
    :param header_name: Optional[str] header name of lif. Does nothing image file is a tif
    :param verbose: bool print status of image loading to standard out.
    :return: np.array image matrix from file, aborts if the image is too large and returns None.
    """

    image_base = None


    if verbose:
        print(f'Loading: {file}...')

    if not os.path.exists(file):
        print(f'\n\x1b[1;31;40m' + f'Cannot access: \'{file}\'. No such file.' + '\x1b[0m')
        return None

    if file.endswith('.lif'):  # Load lif file
        reader = Reader(file)
        series = reader.getSeries()
        for i, header in enumerate(reader.getSeriesHeaders()):
            if header.getName() == header_name:  ###'TileScan 1 Merged':


                chosen = series[i]
                for c in range(4):
                    if c == 0:
                        image_base = chosen.getXYZ(T=0, channel=c)[np.newaxis]

                        if image_base.size * 4 > 9000 * 9000 * 40 * 4:
                            print(f'\x1b[1;33;40m' + f'WARNING: \'{file}\' has image size of {image_base.shape} and is very large. Analysis may fail.' + '\x1b[0m')
                            return None

                    else:
                        image_base = np.concatenate((image_base, chosen.getXYZ(T=0, channel=c)[np.newaxis]), axis=0)
                del series, header, chosen

    elif file.endswith('.tif'): # Load a tif
        image_base = io.imread(file)

        if image_base.ndim == 4:
            image_base = image_base.transpose((-1, 1, 2, 0))

        elif image_base.ndim == 3 and np.array(image_base.shape).min() > 4:  # Suppose you load a 3D image with one channel.
            image_base = image_base[np.newaxis, ...]
            image_base = np.concatenate((image_base, image_base, image_base, image_base), axis=0).transpose(0, 2, 3, 1)
        elif image_base.ndim == 3 and image_base.shape[-1] <= 4:  # Suppose you load a 2D image! with multiple channels
            image_base = image_base[..., [2, 3]].transpose((2, 0, 1)) # always just take the last two channels
            # just return it...
            pass
        elif image_base.ndim == 3 and image_base.shape[0] <= 4:  # Suppose you load a 2D image! with multiple channels
            pass
        else:
            print(f'\x1b[1;31;40m' + f'Cannot load: \'{file}\'. Unsupported number of dimmensions: {image_base.ndim}' + '\x1b[0m')
            return None

    else:
        print(f'\x1b[1;31;40m' + f'Cannot load: \'{file}\'. Filetype not supported.' + '\x1b[0m')
        return None

    return image_base


def correct_pixel_size(image: torch.Tensor, pixel_size: float = 288.88, antialias: Optional[bool] = True):
    """
    Correct an image to have a pixel width of 288.88

    :param image:
    :param pixel_size:
    :param alias:
    :return:
    """

    if pixel_size is None:
        pixel_size = 288.88

    scale = pixel_size / 288.88
    c, x, y = image.shape
    new_size = [round(x * scale), round(y * scale)]
    if image.min() < 0:
        image = image.div(0.5).add(0.5)
        return torchvision.transforms.functional.resize(image, new_size, antialias=antialias).mul(0.5).add(0.5)
    else:
        return torchvision.transforms.functional.resize(image, new_size, antialias=antialias)


def scale_to_hair_cell_diameter(image: torch.Tensor, cell_diameter: int, antialias: Optional[bool] = True) -> torch.Tensor:

    scale = 30 / cell_diameter
    c, x, y = image.shape
    new_size = [round(x * scale), round(y * scale)]
    if image.min() < 0:
        image = image.div(0.5).add(0.5)
        return torchvision.transforms.functional.resize(image, new_size, antialias=antialias).mul(0.5).add(0.5)
    else:
        return torchvision.transforms.functional.resize(image, new_size, antialias=antialias)


def get_dtype_offset(dtype: str = 'uint16') -> int:
    """ get dtype from string """
    if dtype == 'uint16':
        scale = 2 ** 16
    elif dtype == 'uint8':
        scale = 2 ** 8
    elif dtype == 'uint12':  # FM143 usually is this...
        scale = 2 ** 12
    else:
        print(f'\x1b[1;31;40m' + f'ERROR: Unsupported dtype: {dtype}. Currently support: (uint8, uint12, uint16)' + '\x1b[0m')
        scale = None
    return scale


def cochlea_to_xml(cochlea) -> None:
    # Xml header and root
    filename = cochlea.filename
    path = cochlea.path # has filename appended to end
    folder = os.path.split(filename)[0]
    folder = os.path.split(folder)[-1] if len(folder) != 0 else os.path.split(os.getcwd())[1]

    _, height, width= cochlea.im_shape
    depth = 1


    root = ET.Element('annotation')
    ET.SubElement(root, 'folder').text = folder
    ET.SubElement(root, 'filename').text = filename
    ET.SubElement(root, 'path').text = path
    source = ET.SubElement(root, 'source')
    ET.SubElement(source, 'database').text = 'unknown'
    size = ET.SubElement(root, 'size')
    ET.SubElement(size, 'width').text = str(width)
    ET.SubElement(size, 'height').text = str(height)
    ET.SubElement(size, 'depth').text = str(depth)
    ET.SubElement(root, 'segmented').text = '0'


    for c in cochlea.cells:
        x0, y0, x1, y1 = c.boxes
        #  xml write xmin, xmax, ymin, ymax
        object = ET.SubElement(root, 'object')
        ET.SubElement(object, 'name').text = c.type
        ET.SubElement(object, 'pose').text = 'Unspecified'
        ET.SubElement(object, 'truncated').text = '0'
        ET.SubElement(object, 'difficult').text = '0'
        bndbox = ET.SubElement(object, 'bndbox')
        ET.SubElement(bndbox, 'xmin').text = str(int(x0))
        ET.SubElement(bndbox, 'ymin').text = str(int(y0))
        ET.SubElement(bndbox, 'xmax').text = str(int(x1))
        ET.SubElement(bndbox, 'ymax').text = str(int(y1))

    tree = ET.ElementTree(root)
    filename = os.path.splitext(cochlea.path)[0]
    tree.write(filename + '.xml')


########################################################################################################################
#                                                         Plotting
########################################################################################################################

def plot_embedding(embedding: torch.Tensor, centroids: torch.Tensor) -> None:
    num = 25
    x = embedding.detach().cpu().numpy()[0, 0, ...].flatten() * num
    y = embedding.detach().cpu().numpy()[0, 1, ...].flatten() * num
    plt.hist2d(y, x, bins=(embedding.shape[2], embedding.shape[3]))
    print(centroids)
    plt.plot(centroids[0, :, 1].cpu().numpy(), centroids[0, :, 0].cpu().numpy(), 'ro')

    plt.show()

if __name__ == "__main__":
    cochlea = torch.load('test.cochlea')
    cochlea_to_xml(cochlea)


import torch
import torch.nn as nn



class RecurrentUnet(nn.Module):

    def __init__(self, in_channels=4, out_channels=3):
        super(RecurrentUnet, self).__init__()
        channels = [8,16,32,64,128]
        channels = [4,8,16,32,64]

        # self.conv9 = nn.Conv3d(in_channels, channels[0], kernel_size=9, padding=4)
        self.conv7 = nn.Conv3d(in_channels, channels[1], kernel_size=7, padding=3)
        self.conv5 = nn.Conv3d(channels[1], channels[2], kernel_size=5, padding=2)
        self.conv3 = nn.Conv3d(channels[2], channels[3], kernel_size=3, padding=1)

        self.intermediate_conv = nn.Conv3d(channels[3]*2, channels[3], kernel_size=1,padding=0)

        self.fz = nn.Conv3d(channels[3], channels[4], kernel_size=3, padding=1)
        self.fh = nn.Conv3d(channels[3], channels[4], kernel_size=3, padding=1)

        self.up_conv = nn.ConvTranspose3d(channels[3], channels[3], kernel_size=3, stride=2)
        self.out_conv = nn.Conv3d(channels[3], out_channels, 1)

        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()
        self.max_pool = nn.MaxPool3d((2,2,2))


    def forward(self, image):
        # image = self.tanh(self.conv9(image))
        image = self.tanh(self.conv7(image))
        image = self.tanh(self.conv5(image))
        image = self.tanh(self.conv3(image))
        image = self.max_pool(image)

        for t in range(10):
            if t == 0:
                s_t = torch.zeros(image.shape).cuda()

            image = torch.cat((image, s_t), dim=1)
            image = self.intermediate_conv(image)

            # Recurrent Bit!!!
            h = self.tanh(self.fh(image))
            if t == 0:
                h_t = torch.ones(h.shape).cuda()

            z = self.sigmoid(self.fz(image))
            h_t = (h_t * z) + (-1 * z * h)

            s_t = image

        image = self.up_conv(image)


        return self.out_conv(image)



import torch
import torch.nn as nn
import glob
from warnings import filterwarnings


filterwarnings("ignore", category=UserWarning)


class Unet_Constructor(nn.Module):
    def __init__(self,
                 image_dimensions=2,
                 in_channels=3,
                 out_channels=2,
                 feature_sizes=[32, 64, 128, 256, 512, 1024],
                 kernel=(3, 3),
                 upsample_kernel=(2, 2),
                 max_pool_kernel=(2, 2),
                 upsample_stride=2,
                 dilation=1,
                 groups=1,
                 ):
        """
        A Generic Unet Builder Allowing for various different symetric archetectures.
        Generates lists of functions which are looped over in self.forward() to evaluate Unet

        Can pass Dict for each


        :param conv_functions: Tuple of functions from torch.nn
            conv_functions[0]: torch.nn.Conv2d or torch.nn.Conv3d
            conv_functions[1]: torch.nn.ConvTranspose2d or torch.nn.ConvTranspose3d
            conv_functions[2]: torch.nn.MaxPool2d or torch.nn.MaxPool3d

        :param in_channels: Int: Number of input Color Channels; Default is 3
        :param out_channels: Int: Number of classes to predict masks for
        :param kernel: Tuple: Tuple: Convolution Kernel for function in conv_functions[0]
        :param upsample_kernel: Tuple: Convolution Kernel for torch.nn.ConvTranspose2/3d in conv_functions[1]
        :param max_pool_kernel: Tuple: Kernel for torch.nn.MaxPool2/3d
        :param feature_sizes: List: List of integers describing the number of feature channels at each step of the U
        """
        super(Unet_Constructor, self).__init__()
        if image_dimensions == 2:
            conv_functions = (nn.Conv2d, nn.ConvTranspose2d, nn.MaxPool2d, nn.BatchNorm2d)
        elif image_dimensions == 3:
            conv_functions = (nn.Conv3d, nn.ConvTranspose3d, nn.MaxPool3d, nn.BatchNorm3d)
        else:
            raise ValueError(f'Does not support {image_dimensions} dimensional images')

        # Convert to dict of parameters
        #  In order to allow for multiple values passed to the first and second step of each convolution,
        #  we construct a tuple wit values of 'conv1' and 'conv2 denoting the parameter for each step

        if type(kernel) is tuple:
            kernel = {'conv1': kernel, 'conv2': kernel}
        if type(dilation) is int or type(dilation) is tuple:
            dilation = {'conv1': dilation, 'conv2': dilation}
        if type(groups) is int or type(groups) is tuple:
            groups = {'conv1': groups, 'conv2': groups}

        # Error Checking
        if len(feature_sizes) < 2:
            raise ValueError(f'The Number of Features must be at least 2, not {len(feature_sizes)}')
        for i, f in enumerate(feature_sizes[0:-1:1]):
            assert f*2 == feature_sizes[i+1], \
                f'Feature Sizes must be multiples of two from each other: {f} != {feature_sizes[i-1]}*2'

        # Create Dict for saving model
        self.model_specification = {
            'image_dimensions': image_dimensions,
            'in_channels': in_channels,
            'out_channels': out_channels,
            'feature_sizes': feature_sizes,
            'kernel': kernel,
            'upsample_kernel': upsample_kernel,
            'max_pool_kernel': max_pool_kernel,
            'upsample_stride': upsample_stride,
            'dilation': dilation,
            'groups': groups
                                    }

        self.down_steps = []
        self.up_steps = []

        self.down_steps.append(Down(conv_functions,
                                    in_channels=in_channels,
                                    out_channels=feature_sizes[0],
                                    kernel=kernel,
                                    dilation=dilation,
                                    groups=groups,
                                    ))
        i = 1
        for f in feature_sizes[1::]:
            self.down_steps.append(Down(conv_functions,
                                        in_channels=feature_sizes[i-1],
                                        out_channels=f,
                                        kernel=kernel,
                                        dilation=dilation,
                                        groups=groups,
                                        ))
            i += 1
        i = -2
        for f in feature_sizes[:0:-1]:
            self.up_steps.append(Up(conv_functions,
                                    in_channels=f,
                                    out_channels=feature_sizes[i],
                                    kernel=kernel,
                                    upsample_kernel=upsample_kernel,
                                    upsample_stride=upsample_stride,
                                    dilation=dilation,
                                    groups=groups,
                                    ))
            i += -1

        self.out_conv = conv_functions[0](feature_sizes[0], out_channels, 1)
        self.down_steps = nn.ModuleList(self.down_steps)
        self.up_steps = nn.ModuleList(self.up_steps)
        self.max_pool = conv_functions[2](max_pool_kernel)

    def forward(self, x):
        outputs = []

        for step in self.down_steps[:-1]:
            x = step(x)
            outputs.append(x)
            x = self.max_pool(x)

        x = self.down_steps[-1](x)

        for step in self.up_steps:
            x = step(x, outputs.pop())

        x = self.out_conv(x)

        # outputs = None  # Do this to clear memory
        del outputs

        return x

    def save(self, filename, hyperparameters=None):
        model = {'state_dict': self.state_dict(),
                 'model_specifications': self.model_specification,
                 'hyperparameters': hyperparameters}

        python_files = {}

        python_files_list= glob.glob('./**/*.py', recursive=True)
        for f in glob.glob('./**/*.ipynb', recursive=True):
            python_files_list.append(f)

        for f in python_files_list:
            file = open(f,'r')
            python_files[f] = file.read()
            file.close()

        model['python_files'] = python_files
        model['tree_structure'] = glob.glob('**/*', recursive=True)

        torch.save(model, filename)
        return None

    def load(self, filename, to_cuda=True):

        if torch.cuda.is_available() and to_cuda:
            device = 'cuda:0'
        else:
            Warning('Cuda is not available, initializing model on the CPU')
            device = 'cpu'

        model = torch.load(filename, map_location=device)
        model_specification = model['model_specifications']

        self.__init__(
                 image_dimensions=model_specification['image_dimensions'],
                 in_channels=model_specification['in_channels'],
                 out_channels=model_specification['out_channels'],
                 feature_sizes=model_specification['feature_sizes'],
                 kernel=model_specification['kernel'],
                 upsample_kernel=model_specification['upsample_kernel'],
                 max_pool_kernel=model_specification['max_pool_kernel'],
                 upsample_stride=model_specification['upsample_stride'],
                 dilation=model_specification['dilation'],
                 groups=model_specification['groups'],
                 )

        self.load_state_dict(model['state_dict'])
        self.eval()
        try:
            return model['hyperparameters']
        except KeyError:
            return None

class Down(nn.Module):
    def __init__(self, conv_functions: tuple,
                 in_channels: int,
                 out_channels: int,
                 kernel: dict,
                 dilation: dict,
                 groups: dict,
                 ):
        super(Down, self).__init__()

        self.conv1 = conv_functions[0](in_channels,
                                       out_channels,
                                       kernel['conv1'],
                                       dilation=dilation['conv1'],
                                       groups=groups['conv1'],
                                       padding=0)
        self.conv2 = conv_functions[0](out_channels,
                                       out_channels,
                                       kernel['conv2'],
                                       dilation=dilation['conv2'],
                                       groups=groups['conv2'],
                                       padding=0)

        self.batch1 = conv_functions[3](out_channels)
        self.batch2 = conv_functions[3](out_channels)
        self.relu = nn.Tanh()

    def forward(self, x):
        x = self.relu(self.batch1(self.conv1(x)))
        x = self.relu(self.batch2(self.conv2(x)))
        return x


class Up(nn.Module):
    def __init__(self, conv_functions: tuple,
                 in_channels: int,
                 out_channels: int,
                 kernel: tuple,
                 upsample_kernel: tuple,
                 upsample_stride: int,
                 dilation: dict,
                 groups: dict,
                 ):

        super(Up, self).__init__()
        self.conv1 = conv_functions[0](in_channels,
                                       out_channels,
                                       kernel['conv1'],
                                       dilation=dilation['conv1'],
                                       groups=groups['conv1'],
                                       padding=0)
        self.conv2 = conv_functions[0](out_channels,
                                       out_channels,
                                       kernel['conv2'],
                                       dilation=dilation['conv2'],
                                       groups=groups['conv2'],
                                       padding=0)
        if conv_functions[1] == torch.nn.modules.conv.ConvTranspose3d:
            self.up_conv = conv_functions[1](in_channels,
                                             out_channels,
                                             upsample_kernel,
                                             stride=upsample_stride,
                                             padding=0)
            self.lin_up = False
        elif conv_functions[1] == torch.nn.Upsample:
            self.lin_up = True
        else:
            raise RuntimeError('fuck', conv_functions[1])

        self.batch1 = conv_functions[3](out_channels)
        self.batch2 = conv_functions[3](out_channels)
        self.relu = nn.Tanh()

    def forward(self, x, y):
        x = self.up_conv(x)
        y = crop(x, y)
        x = torch.cat((x, y), dim=1)
        x = self.relu(self.batch1(self.conv1(x)))
        x = self.relu(self.batch2(self.conv2(x)))
        return x


@torch.jit.script
def crop(x, y):
    """
    Cropping Function to crop tensors to each other. By default only crops last 2 (in 2d) or 3 (in 3d) dimensions of
    a tensor.

    :param x: Tensor to be cropped
    :param y: Tensor by who's dimmension will crop x
    :return:
    """
    shape_x = x.shape
    shape_y = y.shape
    cropped_tensor = torch.empty(0)

    assert shape_x[1] == shape_y[1],\
        f'Inputs do not have same number of feature dimmensions: {shape_x} | {shape_y}'

    if len(shape_x) == 4:
        cropped_tensor = x[:, :, 0:shape_y[2]:1, 0:shape_y[3]:1]
    if len(shape_x) == 5:
        cropped_tensor = x[:, :, 0:shape_y[2]:1, 0:shape_y[3]:1, 0:shape_y[4]:1]

    return cropped_tensor







import hcat
import hcat.lib.utils as utils
import hcat.lib.functional as functional
from hcat.lib.utils import calculate_indexes, remove_edge_cells, remove_wrong_sized_cells, graceful_exit
from hcat.lib.cell import Cell
from hcat.lib.cochlea import Cochlea
from hcat.backends.spatial_embedding import SpatialEmbedding
from hcat.backends.unet_and_watershed import UNetWatershed

import torch
from torch import Tensor
from tqdm import tqdm
from itertools import product
import numpy as np
from hcat.lib.explore_lif import get_xml

import os.path
import gc
import psutil
from typing import Optional, List, Dict


# DOCUMENTED

def _segment(f: str, channel: int, intensity_reject_threshold: float,
             dtype: str = 'uint16',
             unet: bool = False, cellpose: bool = False, no_post: bool = False,
             figure: Optional[bool] = True) -> Cochlea:
    """

    Segment arbitrarily large images of cochleae.
    Loads a lif or tif file and iterativley segments all hair cells in the image.
    Saves the results of the analysis in a file with extension ".cochlea" which compresses and saves the analysis.

    Example:

    >>> from hcat.segment import _segment
    >>> f = 'path/to/my/file.tif'
    >>> segment(f, intensity_reject_threshold=0.5)

    :param *str* f: path to file
    :param *int* channel: index of cytosolic stain (0 if only one channel).
    :param *float* intensity_reject_threshold: predicted cell masks with mean cytosolic stain below this threshold will be rejected
    :param *str,optional* dtype: bit depth of input image. Possible values: uint8, uint16, int8, int16.
    :param *bool,optional* unet: Perform segmentation with UNet + Watershed segmentation backbone.
    :param *bool,optinal* cellpose: Perform segmentation with 3D cellpose segmentation backbone.
    :param *bool,optinal* no_post: if true, remove cell postprocessing, which may remove legitimate cells.
    :param *bool,optinal* figure: If true, renders a diagnostic figure for this cochlea.
    :return: None
    """
    print('Initializing hair cell segmentation algorithm...')
    if f is None:
        print('\x1b[1;31;40m' + 'ERROR: No File to Analyze... \nAborting.' + '\x1b[0m')
        return None

    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    if device == 'cuda':
        print('\x1b[1;32;40m' + 'CUDA: GPU successfully initialized!' + '\x1b[0m')
    else:
        print(
            '\x1b[1;33;40m' + 'WARNING: GPU not present or CUDA is not correctly intialized for GPU accelerated computation. '
                              'Analysis may be slow.' + '\x1b[0m')

    max_cell: int = 0
    cells = []
    append_cells = cells.append
    intensity_reject: callable = torch.jit.script(
        hcat.lib.functional.IntensityCellReject(threshold=intensity_reject_threshold).requires_grad_(False))
    predict_curvature: callable = hcat.lib.functional.PredictCurvature(erode=3)

    with torch.no_grad():

        backend = _get_backend(unet, cellpose, device)  # may return none if failed
        if backend is None:
            return None

        image_base = utils.load(f, 'TileScan 1 Merged', verbose=True)  # may return none if failed
        print('SEG SHAPE', image_base.shape)
        if image_base is None:
            return None

        print(
            f'DONE: shape: {image_base.shape}, min: {image_base.min()}, max: {image_base.max()}, dtype: {image_base.dtype}')

        dtype = image_base.dtype if dtype is None else dtype
        scale: int = hcat.lib.utils.get_dtype_offset(dtype)

        c, x, y, z = image_base.shape
        out_img = torch.zeros((1, x, y, z), dtype=torch.long)  # (1, X, Y, Z)

        x_ind: List[List[int]] = calculate_indexes(15, 255, x, x)  # [[0, 255], [30, 285], ...]
        y_ind: List[List[int]] = calculate_indexes(15, 255, y, y)  # [[0, 255], [30, 285], ...]
        total: int = len(x_ind) * len(y_ind)

        for (x, y) in tqdm(product(x_ind, y_ind), total=total, desc='Segmenting: '):

            # Load and prepare image for ML model evaluation
            image: np.array = image_base[:, x[0]:x[1], y[0]:y[1]].astype(np.float)
            image: Tensor = torch.from_numpy(image).float().div(scale).unsqueeze(0)
            image: Tensor = image.div(0.5) if image.min() < 0 else image.sub(0.5).div(0.5)  # signed or unsigned?

            # which backend do we use?
            out: Tensor = backend(image.to(device)) if unet else backend(image[:, [channel], ...].to(device))

            # backend will return an empty tensor if there wasn't usable data underneath
            if out.shape[1] == 0:
                continue

            # postprocess
            out: Tensor = intensity_reject(out, image[:, [channel], ...].to(device)) if not no_post else out

            # backend will return an empty tensor if there wasn't usable data underneath
            if out.shape[1] == 0:
                continue


            value, out = out.max(1)
            out[value < 0.25] = 0

            # postprocess
            out: Tensor = remove_wrong_sized_cells(out) if not no_post else out

            # Remove cells on the edge for proper merging
            out: Tensor = remove_edge_cells(out).cpu()

            if out.shape[1] == 0:
                continue

            out[out > 0] = out[out > 0] + max_cell
            max_cell = out.max() if out.max() > 0 else max_cell

            # merging!
            x_exp = torch.tensor([x[0], x[1]])
            y_exp = torch.tensor([y[0], y[1]])
            x_exp[x_exp < 0] = 0
            y_exp[y_exp < 0] = 0
            destination = out_img[:, x_exp[0]:x_exp[1], y_exp[0]:y_exp[1], :]
            destination, new_ids = functional.merge_regions(destination.to(out.device), out, 0.35)

            out_img[:, x_exp[0]:x_exp[1], y_exp[0]:y_exp[1], :] = destination.cpu()

            for u in new_ids:
                if u == 0:
                    continue
                center = (out == u).nonzero(as_tuple=False).float().mean(0)
                center[1] += x[0]
                center[2] += y[0]
                append_cells(Cell(id=u,
                                  image=torch.clone(image[:, :, :-1:, :-1:, 0:out.shape[-1]].mul(0.5).add(0.5)),
                                  mask=(out == u).unsqueeze(0),
                                  loc=center))
        print('Predicted Number of Cells: ', len(cells))

        mem = psutil.virtual_memory()  # get memory for later
        del image_base  # memory issues are tight. removing image and image_base help
        gc.collect()

        # Not every input will be a whole cochlea... curvature means nothing over such a small region
        curvature, distance, apex = predict_curvature(out_img.sum(-1).unsqueeze(-1), cells, None)

        if curvature is None:
            print('\x1b[1;33;40mWARNING: ' +
                  'All three methods to predict hair cell path have failed. Frequency Mapping functionality is limited.'
                  ' Consider Manual Calculation.'
                  + '\x1b[0m')
        # curvature estimation really only works if there is a lot of tissue...
        if distance is not None and distance.max() > 4000:
            for c in cells: c.calculate_frequency(curvature[[0, 1], :], distance)  # calculate cell's best frequency
            # cells = [c for c in cells if not c._distance_is_far_away]  # remove a cell if its far away from curve
        else:
            curvature, distance, apex = None, None, None
            print('\x1b[1;33;40mWARNING: ' +
                  'Predicted Cochlear Distance is below 4000um. Not sufficient information to determine cell frequency.'
                  + '\x1b[0m')

        xml = get_xml(f) if f.endswith('.lif') else None

        # Store
        c = Cochlea(mask=out_img,
                    filename=f,
                    leica_metadata=xml,
                    analysis_type='segment',
                    im_shape=out_img.shape,
                    cochlear_distance=distance,
                    curvature=curvature,
                    cells=cells,
                    apex=apex)

        # This takes too much memory for huge images
        # Check the size of the outimg. If we have twice the available free memory we can compress!
        if mem.available > (out_img.numpy().size * out_img.numpy().itemsize * 2):
            c.compress_mask(vebose=True)
        else:
            print('\x1b[1;33;40m' + 'WARNING: Low free system memory. Cannot compress output... ' + '\x1b[0m')

        c.save(os.path.splitext(f)[0] + '.cochlea')
        print('Saved: ', os.path.splitext(f)[0] + '.cochlea')

        if figure:
            c.make_fig(os.path.splitext(f)[0] + '.pdf')

        # c.write_csv(os.path.splitext(f)[0] + '.csv')
        c.render_mask(os.path.splitext(f)[0] + '_mask.tif')
        print('')
        return c


@graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not load segmentation backend.' + '\x1b[0m')
def _get_backend(unet: bool, cellpose: bool, device: str):
    """ Get the appropriate Backend """
    if unet and not cellpose:
        backend = UNetWatershed(unet_model_path='/media/DataStorage/Dropbox (Partners HealthCare)/'
                                                'HairCellInstance/checkpoints/'
                                                'May11_16-38-12_chris-MS-7C37_checkpoint.hcnet',
                                frcnn_model_path='/media/DataStorage/Dropbox (Partners HealthCare)/HcUnet/TrainedModels/'
                                                 'fasterrcnn_Oct14_06:05.pth',
                                device=device).requires_grad_(False)
    elif not unet and not cellpose:
        backend = SpatialEmbedding(model_loc='/media/DataStorage/Dropbox (Partners HealthCare)/HairCellInstance/'
                                             'models_lts/spatial_embedding.trch',
                                   sigma=torch.tensor([0.2, 0.2, 0.2]),
                                   scale=25,
                                   device=device).requires_grad_(False)
    elif cellpose and not unet:
        from hcat.backends.cellpose import Cellpose  # cellpose announces it makes a log file
        backend = Cellpose(device=device)

    return backend




# Have to do this first to avoid circular import. I am a bad programer.
class ShapeError(Exception):
    pass

from hcat.backends.spatial_embedding import SpatialEmbedding
from hcat.backends.unet_and_watershed import UNetWatershed




import skimage.scripts.skivi
import torch
from torch import Tensor
from torch.utils.data import DataLoader
import numpy as np
import glob
import os.path
import shutil
from tqdm import tqdm
import skimage.io as io
from typing import Dict, List, Union
from skimage.morphology import ball
import xml

import hcat.train.transforms as t


class dataset(DataLoader):
    def __init__(self, path, transforms=None, random_ind: bool = False, min_cell_volume: int = 4000,
                 duplicate_poor_data: bool = False):
        super(DataLoader, self).__init__()

        # Functions to correct for Bad Masks
        self.colormask_to_mask = t.colormask_to_mask()
        self.transformation_correction = t.transformation_correction(min_cell_volume)  # removes bad masks

        # Reassigning variables
        self.mask = []
        self.image = []
        self.centroids = []
        self.transforms = transforms
        self.dpd = duplicate_poor_data

        # Find only files
        self.files = glob.glob(os.path.join(path, '*.labels.tif'))
        for f in self.files:
            image_path = os.path.splitext(f)[0]
            image_path = os.path.splitext(image_path)[0] + '.tif'

            if image_path.find(
                    '-DUPLICATE') > 0 and not self.dpd:  # If -DUPLICATE flag is there and we want extra copies continue
                continue

            image = torch.from_numpy(io.imread(image_path).astype(np.uint16) / 2 ** 16).unsqueeze(0)
            image = image.transpose(1, 3).transpose(0, -1).squeeze()[[0, 2, 3], ...]

            mask = torch.from_numpy(io.imread(f)).transpose(0, 2).unsqueeze(0)

            self.mask.append(mask.float())
            self.image.append(image.float())
            self.centroids.append(torch.tensor([0, 0, 0]))

        # implement random permutations of the indexing
        self.random_ind = random_ind

        if self.random_ind:
            self.index = torch.randperm(len(self.mask))
        else:
            self.index = torch.arange((len(self.mask)))

    def __len__(self) -> int:
        return len(self.mask)

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:

        with torch.no_grad():
            item = self.index[item]
            data_dict = {'image': self.image[item], 'masks': self.mask[item], 'centroids': self.centroids[item]}
            did_we_get_an_output = False

            while not did_we_get_an_output:
                try:
                    if self.transforms is not None:
                        data_dict = self.transforms(data_dict)
                        did_we_get_an_output = True
                except RuntimeError:
                    continue

            data_dict = self.colormask_to_mask(data_dict)
            assert torch.isnan(data_dict['centroids']).sum() == 0
            data_dict = self.transformation_correction(data_dict)  # removes bad masks

        return data_dict

    def step(self) -> None:
        if self.random_ind:
            self.index = torch.randperm(len(self.mask))


class nul_dataset(DataLoader):
    def __init__(self, path, transforms=None, random_ind: bool = False):

        super(DataLoader, self).__init__()

        # Find only files
        files = glob.glob(os.path.join(path, '*.tif'))

        self.image = []
        self.centroids = []
        self.transforms = transforms

        for f in files:
            image = torch.from_numpy(io.imread(f).astype(np.uint16) / 2 ** 16).unsqueeze(0)
            image = image.transpose(1, 3).transpose(0, -1).squeeze()[[0, 2, 3], ...]
            self.centroids.append(torch.tensor([0, 0, 0]))
            self.image.append(image.float())

        # implement random permutations of the indexing
        self.random_ind = random_ind

        if self.random_ind:
            self.index = torch.randperm(len(self.image))
        else:
            self.index = torch.arange((len(self.image)))

    def __len__(self) -> int:
        return len(self.mask)

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:

        item = self.index[item]
        shape = self.image[item].shape
        shape[0] = 1  # override the channel shape (4 -> 1)

        data_dict = {'image': self.image[item], 'masks': torch.zeros(shape), 'centroids': self.centroids[item]}
        did_we_get_an_output = False

        while not did_we_get_an_output:
            try:
                if self.transforms is not None:
                    data_dict = self.transforms(data_dict)
                    did_we_get_an_output = True
            except RuntimeError:
                continue

        return data_dict

    def step(self) -> None:
        if self.random_ind:
            self.index = torch.randperm(len(self.image))


class em_dataset(DataLoader):
    def __init__(self, path, transforms=None, random_ind: bool = False, big_path =None):
        super(DataLoader, self).__init__()

        # Find only files
        files = glob.glob(os.path.join(path, '*.labels.tif'))

        self.mask = []
        self.image = []
        self.centroids = []
        self.transforms = transforms

        # Functions to correct for Bad Masks
        self.colormask_to_mask = t.colormask_to_mask()
        self.transformation_correction = t.transformation_correction(4000)  # removes bad masks


        for f in files:
            image_path = os.path.splitext(f)[0]
            image_path = os.path.splitext(image_path)[0] + '.tif'
            image = torch.from_numpy(io.imread(image_path).astype(np.uint16) / 2 ** 8).unsqueeze(-1)
            if image.shape[-1] > 1:
                image = image.transpose(0, -1).transpose(1,2).squeeze()[[0, 2, 3], ...]
            else:
                image = image.transpose(0, -1).transpose(1,2).squeeze().unsqueeze(0)

            mask = torch.from_numpy(io.imread(f)).transpose(0, 2).unsqueeze(0)


            self.mask.append(mask.float())
            self.image.append(image.float())
            self.centroids.append(torch.tensor([0]))


        if big_path:
            if not isinstance(big_path, list):
                big_path = [big_path]

            for path in big_path:
                stacks = glob.glob(path+'/mito_train/*.tif')
                imgs = glob.glob(path+'/im/*.png')
                stacks.sort()
                imgs.sort()

                big_img = torch.zeros((4096, 4096, len(stacks)), dtype=torch.uint8)
                big_mask = torch.zeros(big_img.shape, dtype=torch.int16)
                for i, s in tqdm(enumerate(stacks), total=len(stacks)):
                    m = io.imread(s)
                    big_mask[:,:,i] = torch.from_numpy(m.astype(np.int32)).type(torch.int16).mul_(-1).add_(2**16)

                    m = io.imread(imgs[i])
                    big_img[:,:,i]  = torch.from_numpy(m.astype(np.int32)).type(torch.int16)

                self.mask.append(big_mask.float().unsqueeze(0))
                self.image.append(big_img.float().unsqueeze(0))
                self.centroids.append(torch.tensor([0]))

                del big_mask, big_img

        # implement random permutations of the indexing
        self.random_ind = random_ind

        if self.random_ind:
            self.index = torch.randperm(len(self.mask))
        else:
            self.index = torch.arange((len(self.mask)))

    def __len__(self) -> int:
        return len(self.mask)


    def step(self) -> None:
        if self.random_ind:
            self.index = torch.randperm(len(self.mask))

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:
        with torch.no_grad():
            item = self.index[item]
            data_dict = {'image': self.image[item], 'masks': self.mask[item], 'centroids': self.centroids[item]}
            did_we_get_an_output = False
            i = 0
            while not did_we_get_an_output:
                i += 1
                if self.transforms is not None:
                    data_dict = self.transforms(data_dict)
                    did_we_get_an_output = True

            data_dict = self.colormask_to_mask(data_dict)
            assert torch.isnan(data_dict['centroids']).sum() == 0
            data_dict = self.transformation_correction(data_dict)  # removes bad masks


        return data_dict


class synthetic:
    def __init__(self):
        pass

    def __len__(self):
        return 1

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:
        ball = torch.from_numpy(skimage.morphology.ball(15))
        mask = torch.zeros((1, 156, 156, 31))
        image = torch.zeros((4, 156, 156, 31))

        x = torch.randint(156 - 32, (1,))
        y = torch.randint(156 - 32, (1,))

        mask[0, x:x + 31, y:y + 31, :] = ball
        image[1, x:x + 31, y:y + 31, :] = ball
        image = (2 * image / 3) + (torch.rand((4, 156, 156, 31)) / 3)
        centroids = torch.zeros((1, 3))
        centroids[0, :] = torch.tensor([x + 31 // 2, y + 31 // 2, 31 // 2])

        return {'masks': mask.cuda(), 'image': image.cuda(), 'centroids': centroids.cuda()}

    def step(self):
        pass


class detection_dataset(DataLoader):
    def __init__(self, path, transforms=None, random_ind: bool = False, min_cell_volume: int = 20*10,
                 duplicate_poor_data: bool = False, simple_class: bool = True, batch_size: int = 1):
        super(DataLoader, self).__init__()

        # Functions to correct for Bad Masks
        self.transformation_correction = t.transformation_correction(min_cell_volume)  # removes bad masks

        # Reassigning variables
        self.mask = []
        self.image = []
        self.centroids = []
        self.boxes = []
        self.labels = []
        self.transforms = transforms
        self.dpd = duplicate_poor_data

        # Find only files
        self.simple_class = simple_class
        self.files = glob.glob(f'{path}{os.sep}*.xml')

        for f in self.files:
            image_path = f[:-4:]+ '.tif'
            # print(image_path, '\n', f)
            # image_path = f

            if image_path.find(
                    '-DUPLICATE') > 0 and not self.dpd:  # If -DUPLICATE flag is there and we want extra copies continue
                continue

            # [c, X, Y] -> [c, x, y, z=1]
            image = torch.from_numpy(io.imread(image_path).astype(np.uint16) / 2 ** 16).unsqueeze(-1)
            mask = torch.empty((0,0,0,0))

            self.mask.append(mask.float())
            self.image.append(image.float())
            self.centroids.append(torch.tensor([0, 0, 0]))

            tree = xml.etree.ElementTree.parse(f)
            root = tree.getroot()
            bbox_loc = []
            class_labels = []

            for c in root.iter('object'):
                for cls in c.iter('name'):
                    class_labels.append(cls.text)

                for a in c.iter('bndbox'):
                    x1 = int(a.find('xmin').text)
                    y1 = int(a.find('ymin').text)
                    x2 = int(a.find('xmax').text)
                    y2 = int(a.find('ymax').text)
                    bbox_loc.append([x1, y1, x2, y2])

            for i, s in enumerate(class_labels):
                if s == 'OHC':
                    class_labels[i] = 1
                elif s == 'OHC1':
                    class_labels[i] = 1
                elif s == 'OHC2':
                    class_labels[i] = 2
                elif s == 'OHC3':
                    class_labels[i] = 3
                elif s == 'IHC':
                    class_labels[i] = 4
                else:
                    print(class_labels)
                    print(bbox_loc)
                    raise ValueError('Unidentified Label in XML file of ' + f)

            class_labels = torch.tensor(class_labels)

            # For testing - reduce row specific labels in cell specific (IHC or OHC)
            if self.simple_class:
                class_labels[class_labels == 2] = 1  # OHC2 -> OHC
                class_labels[class_labels == 3] = 1  # OHC3 -> OHC
                class_labels[class_labels == 4] = 2  # IHC

            self.boxes.append(torch.tensor(bbox_loc))
            self.labels.append(class_labels)

        # implement random permutations of the indexing
        self.random_ind = random_ind
        self.batch_size = batch_size

        self.index = self._batch_ind(batch_size=self.batch_size, dataset_len=len(self.image), random_ind=self.random_ind)

    @staticmethod
    def _batch_ind( batch_size: int, dataset_len: int, random_ind: bool = False) -> List[List[int]]:
        """

        :param batch_size:
        :param dataset_len:
        :return:
        """
        if dataset_len % batch_size >= batch_size//2:
            end = dataset_len + dataset_len % batch_size
        else:
            end = dataset_len + batch_size - (dataset_len % batch_size)

        base_ind: Tensor = torch.arange(0, end) # should make evenly

        if not base_ind.shape[0] % batch_size == 0:
            print(base_ind.shape, batch_size, dataset_len, dataset_len % batch_size, base_ind.shape[0] % batch_size)
            raise ValueError

        indicies = []
        _ind = []
        for i, ind in enumerate(base_ind):
            if ind >= dataset_len:
                base_ind[i] -= dataset_len

        # Shuffle the indicies if we do a random permutation
        if random_ind: base_ind = base_ind[torch.randperm(base_ind.shape[0])]

        for i, ind in enumerate(base_ind):
            _ind.append(base_ind[i].item())
            if len(_ind) == batch_size:
                indicies.append(_ind)
                _ind = []

        return indicies


    def __len__(self) -> int:
        return len(self.index)

    def __getitem__(self, item: int) -> Union[Dict[str, torch.Tensor], Dict[str, Union[List[torch.Tensor], torch.Tensor]]]:

        if self.batch_size == 1:
            with torch.no_grad():
                item: List[List[int]] = self.index[item]
                item = item[0] # cuz new function

                did_we_get_an_output = False
                i = 0
                while not did_we_get_an_output:
                    if self.transforms is not None:
                        data_dict = {'image': self.image[item],
                                     'masks': self.mask[item],
                                     'centroids': self.centroids[item],
                                     'boxes': self.boxes[item],
                                     'labels': self.labels[item]}

                        # Transformation pipeline
                        data_dict = t.box_to_mask()(data_dict)                 # For geometric transforms
                        data_dict = self.transforms(data_dict)                 # Apply transforms
                        data_dict = self.transformation_correction(data_dict)  # removes bad masks/boxes
                        data_dict = t.mask_to_box()(data_dict)


                        # may not be boxes remaining, try again if...
                        if data_dict['boxes'].shape[0] > 0:
                            did_we_get_an_output = True
                        else:
                            i += 1
                            data_dict['masks'] = None
                    if i > 10:
                        print('MOVING')
                        shutil.move(self.files[item],'/home/chris/Desktop/')
                        """ Sometimes gets stuck in infinite loop - RARE """
                        raise ValueError

                    # some transforms throw an error if random things get bad. Most shouldn't. Just ignore and try again.



            data_dict['boxes'] = torch.round(data_dict['boxes'].float())
            c, x, y, z = data_dict['image'].shape
            data_dict['image'] = torch.cat((torch.zeros((1, x, y, z), device=data_dict['image'].device),
                                            data_dict['image']), dim=0)
            data_dict['image'] = data_dict['image'][:, :, :, 0]
            data_dict.pop('masks', None)
            return data_dict

        else:  # if batch size > 1
            out = []

            with torch.no_grad():
                item_list: List[List[int]] = self.index[item]

                for item in item_list:
                    did_we_get_an_output = False
                    i = 0
                    while not did_we_get_an_output:
                        if self.transforms is not None:
                            data_dict = {'image': self.image[item],
                                         'masks': self.mask[item],
                                         'centroids': self.centroids[item],
                                         'boxes': self.boxes[item],
                                         'labels': self.labels[item]}

                            # Transformation pipeline
                            data_dict = t.box_to_mask()(data_dict)                 # For geometric transforms
                            data_dict = self.transforms(data_dict)                 # Apply transforms
                            data_dict = self.transformation_correction(data_dict)  # removes bad masks/boxes
                            data_dict = t.mask_to_box()(data_dict)


                            # may not be boxes remaining, try again if...
                            if data_dict['boxes'].shape[0] > 0:
                                did_we_get_an_output = True
                            else:
                                i += 1
                                data_dict['masks'] = None
                        if i > 10:
                            print('MOVING')
                            shutil.move(self.files[item],'/home/chris/Desktop/')
                            """ Sometimes gets stuck in infinite loop - RARE """
                            raise ValueError

                        # some transforms throw an error if random things get bad. Most shouldn't. Just ignore and try again.


                data_dict['boxes'] = torch.round(data_dict['boxes'].float())
                c, x, y, z = data_dict['image'].shape
                data_dict['image'] = torch.cat((torch.zeros((1, x, y, z), device=data_dict['image'].device),
                                                data_dict['image']), dim=0)
                data_dict['image'] = data_dict['image'][:, :, :, 0]
                data_dict.pop('masks', None)
                out.append(data_dict)

            return out  # List[Dict[str, Tensor]]

    def step(self) -> None:
        if self.random_ind:
            self.index = self._batch_ind(batch_size=self.batch_size, dataset_len=len(self.image), random_ind=self.random_ind)


@torch.jit.script
def colormask_to_torch_mask(colormask: torch.Tensor) -> torch.Tensor:
    """

    :param colormask: [C=1, X, Y, Z]
    :return:
    """
    uni = torch.unique(colormask)
    uni = uni[uni != 0]
    num_cells = len(uni)

    c, x, y, z = colormask.shape

    shape = (num_cells, x, y, z)
    mask = torch.zeros(shape)

    for i, u in enumerate(uni):
        mask[i, :, :, :] = (colormask[0, :, :, :] == u)

    return mask


@torch.jit.script
def colormask_to_centroids(colormask: torch.Tensor) -> torch.Tensor:
    uni = torch.unique(colormask)
    uni = uni[uni != 0]
    num_cells = len(uni)  # cells are denoted by integers 1->max_cell
    shape = (num_cells, 3)
    centroid = torch.zeros(shape)

    for i, u in enumerate(uni):
        indexes = torch.nonzero(colormask[0, :, :, :] == u).float()
        centroid[i, :] = torch.mean(indexes, dim=0)

    # centroid[:, 0] /= colormask.shape[1]
    # centroid[:, 1] /= colormask.shape[2]
    # centroid[:, 2] /= colormask.shape[3]

    return centroid


if __name__ == "__main__":
    data = synthetic()

    out = data[1]
    print(out['centroids'].shape)



import torch.nn as nn
import torch
from hcat.train.transforms import _crop
from typing import List


class r_unet(nn.Module):
    """ base class for spatial embedding and unet """
    def __init__(self, in_channels:  int = 1, n: int = 5, c: List[int] = [20, 25, 30]):
        super(r_unet, self).__init__()

        self.activation = nn.LeakyReLU()

        # c = [15, 40, 80]
        # c = [20, 25, 30]
        # c = [10, 20, 30]

        self.down_1 = r_block(in_channels, c[0], n, 1, 1)
        self.down_2 = r_block(c[0], c[1], n, 1, 1)
        self.down_3 = r_block(c[1], c[2], n, 1, 1)

        self.up_1 = r_block(c[2] + c[1], c[1], n)
        self.up_2 = r_block(c[1] + c[0], c[0], n)

        self.out_embed = nn.Conv3d(c[0], 3, kernel_size=3, stride=1, dilation=1, padding=1)
        self.out_prob = nn.Conv3d(c[0], 1, kernel_size=3, stride=1, dilation=1, padding=1)

        self.stride_1 = nn.Conv3d(c[0], c[0], kernel_size=3, stride=(2, 2, 2), dilation=1, padding=0)
        self.stride_2 = nn.Conv3d(c[1], c[1], kernel_size=3, stride=(2, 2, 2), dilation=1, padding=0)

        self.transpose_1 = nn.ConvTranspose3d(c[2], c[2], kernel_size=3, stride=(2, 2, 2), padding=0, dilation=1)
        self.transpose_2 = nn.ConvTranspose3d(c[1], c[1], kernel_size=3, stride=(2, 2, 2), padding=0, dilation=1)

        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()
        self.softmax = nn.Softmax()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


class embed_model(r_unet):
    """ for spatial embedding """
    def __init__(self, in_channels: int = 1, n=5, c=[20, 25, 30]):
        self.n = n
        self.c = c
        self.in_channels = in_channels

        super(embed_model, self).__init__(in_channels=in_channels, n=n, c=c)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # print(f'Step in: {x.shape}')
        x = self.down_1(x)
        # print(f'Step down1: {x.shape}: {self.stride_1(x).shape}')
        y = self.down_2(self.activation(self.stride_1(x)))
        # print(f'Step down2: {y.shape}')
        z = self.down_3(self.activation(self.stride_2(y)))
        # print(f'Step down3: {z.shape}')
        z = self.activation(self.transpose_1(z))
        # print(f'Step transpose1: {z.shape}')
        z = self.up_1(torch.cat((crop(y, z.shape), z), dim=1))
        # print(f'Step up1: {z.shape}')
        z = self.activation(self.transpose_2(z))
        # print(f'Step transpose2: {z.shape}')
        z = self.up_2(torch.cat((crop(x, z.shape), z), dim=1))
        # print(f'Step up2: {z.shape}')

        z = torch.cat((
            self.tanh(self.out_embed(z)),
            self.sigmoid(self.out_prob(z))
        ), dim=1)
        # print(f'Step out: {z.shape}')
        # raise ValueError

        return z


class wtrshd_model(r_unet):
    """ for watershed segmentation """
    def __init__(self, in_channels: int = 4):
        super(wtrshd_model, self).__init__(in_channels=in_channels)
        self.out_embed = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.down_1(x)
        y = self.down_2(self.activation(self.stride_1(x)))
        z = self.down_3(self.activation(self.stride_2(y)))
        z = self.activation(self.transpose_1(z))
        z = self.up_1(torch.cat((crop(y, z.shape), z), dim=1))
        z = self.activation(self.transpose_2(z))
        z = self.up_2(torch.cat((crop(x, z.shape), z), dim=1))

        z = self.sigmoid(self.out_prob(z))

        return z


class r_block(nn.Module):
    """
    Recurrently applies a block to an input and outputs later.
    """
    def __init__(self, in_channels: int, out_channels: int, n_loop: int, dilation: int = 1, padding: int = 1):
        super(r_block, self).__init__()

        self.n = n_loop

        self.conv1x1 = nn.Conv3d(in_channels * 2, in_channels, kernel_size=1)
        # self.conv7x7_3 = nn.Conv3d(in_channels, in_channels, kernel_size=(3,3,1), padding=(1,1,0), stride=1, dilation=1)
        self.conv7x7_1 = nn.Conv3d(in_channels, in_channels, kernel_size=3, padding=padding, stride=1, dilation=dilation)
        self.conv7x7_2 = nn.Conv3d(in_channels, out_channels, kernel_size=3, padding=padding, stride=1, dilation=dilation)

        self.activation = nn.LeakyReLU()

        self.batch_norm_conv_1x1 = nn.BatchNorm3d(in_channels)
        self.batch_norm_conv_7x7_1 = nn.BatchNorm3d(in_channels)
        # self.batch_norm_conv_7x7_3 = nn.BatchNorm3d(in_channels)
        self.batch_norm_conv_7x7_2 = nn.BatchNorm3d(out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = torch.zeros(x.shape, device=x.device)
        for _ in range(self.n):
            y = torch.cat((x, y), dim=1)
            y = self.activation(self.batch_norm_conv_1x1(self.conv1x1(y)))
            # y = self.activation(self.batch_norm_conv_7x7_3(self.conv7x7_3(y)))
            y = self.activation(self.batch_norm_conv_7x7_1(self.conv7x7_1(y)))

        y = self.activation(self.batch_norm_conv_7x7_2(self.conv7x7_2(y)))

        return y


@torch.jit.script
def crop(x: torch.Tensor, shape: List[int]) -> torch.Tensor:
    return _crop(x, 0, 0, 0, int(shape[2]), int(shape[3]), int(shape[4]))



from __future__ import print_function
from __future__ import division
import torch
import torchvision


def rcnn(path=None):
    """
    Simple function to return faster rcnn model
    :param path: Path to pretrained model file.
    :return: faster_rcnn model
    """

    model = torchvision.models.detection.fasterrcnn_resnet50_fpn(pretrained=False,
                                                                 progress=True,
                                                                 num_classes=3,
                                                                 pretrained_backbone=True,
                                                                 box_detections_per_img=500)
    if path is not None:
        model.load_state_dict(torch.load(path))
    return model




import torch
import torch.nn as nn
from typing import Dict, Optional

import hcat.lib.functional
from hcat.models.r_unet import wtrshd_model as WatershedUnet
from hcat.models.depreciated.rcnn import rcnn


# DOCUMENTED

class UNetWatershed(nn.Module):
    def __init__(self,
                 device: str = 'cuda',
                 unet_model_path: str = None,
                 frcnn_model_path: str = None):
        """
        Initalize UNet + Watershed Segmentation Algorithm

        :param device: device to run analysis on.
        :param unet_model_path: path to trained unet
        :param frcnn_model_path: path to trained fasterrcnn model
        """

        super(UNetWatershed, self).__init__()

        self._in_channels = 1
        self.device = device

        self.model = torch.jit.script(WatershedUnet(in_channels=self._in_channels)).to(device)
        self.ffcnn = rcnn().to(device).eval()

        self.predict_cell_candidates = hcat.lib.functional.PredictCellCandidates(self.ffcnn, device=device)
        self.predict_segmentation_mask = hcat.lib.functional.PredictSemanticMask(model=self.model, device=device)
        self.generate_seed_map = hcat.lib.functional.GenerateSeedMap()

        self.instance_mask_from_prob = hcat.lib.functional.InstanceMaskFromProb()

    # @graceful_exit('\x1b[1;31;40m' + 'ERROR: Unet+Watershed Failed. Aborting...' + '\x1b[0m')
    def forward(self, image: torch.Tensor, channel: Optional[int] = 2) -> torch.Tensor:
        """
        Inputs an image and outputs a probability mask of everything seen in the image.

        .. note::
           Call the module as a function to execute this method (similar to torch.nn.module).

        .. warning:
           Will not raise an error upon failure, instead returns None and prints to standard out

        Example:

        >>> from hcat.backends.spatial_embedding import SpatialEmbedding
        >>> import torch
        >>> backend = SpatialEmbedding()
        >>> image = torch.load('path/to/my/image.trch')
        >>> assert image.ndim == 5 # Shape should be [B, C, X, Y, Z]
        >>> masks = backend(image)

        :param image: [B, C=4, X, Y, Z] input image
        :return: [B, 1, X, Y, Z] output segmentation mask where each pixel value is a cell id (0 is background)
        """
        assert image.ndim == 5
        if image.shape[1] != 4: raise ValueError('Need 4D image for unet+watershed segmentation.')
        b, c, x, y, z = image.shape

        if self.image_reject and self._is_image_bad(image):
            return torch.zeros((b, 0, x, y, z), device=self.device)

        image_frcnn = image[:, [0, 2, 3], :, :, :]

        cell_candidates: Dict[str, torch.Tensor] = self.predict_cell_candidates(image_frcnn)
        semantic_mask: torch.Tensor = self.predict_segmentation_mask(image[:, [channel], ...])
        seed: torch.Tensor = self.generate_seed_map(cell_candidates, semantic_mask)
        instance_mask: torch.Tensor = self.instance_mask_from_prob(semantic_mask, seed)

        return self._colormask_to_mask(instance_mask)

    def load(self, unet_path: str, fasterrcnn_path: str) -> None:
        """
        Assigns pretrained weights to both necessary backend ML models: unet and fasterrcnn

        :param unet_path: path to pretrained unet model
        :param fasterrcnn_path: path to pretrained faster rcnn model
        :return:
        """

        checkpoint = torch.load(unet_path)
        self.model.load_state_dict(checkpoint['model_state_dict'])
        self.ffcnn = rcnn(path=fasterrcnn_path).to(self.device).eval()


import torch
import torch.nn as nn
from src.models.depreciated.HCBlock import HCBlock
from warnings import filterwarnings


filterwarnings("ignore", category=UserWarning)


class HCNet(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, complexity: int = 30):
        super(HCNet, self).__init__()

        self.conv3x3_1 = nn.Conv3d(in_channels=in_channels, out_channels=10, padding=1, kernel_size=3)
        self.bn_1 = nn.BatchNorm3d(10)

        self.conv5x5_1 = nn.Conv3d(in_channels=10, out_channels=20, kernel_size=5, padding=2)
        self.bn_5x5_1 = nn.BatchNorm3d(20)

        self.strided_conv = nn.Conv3d(20, complexity, kernel_size=3, stride=2, padding=1)
        self.bn_strided = nn.BatchNorm3d(complexity)

        self.hcblock = torch.jit.script(HCBlock(in_channels=complexity))

        self.transposed_conv = nn.ConvTranspose3d(in_channels=complexity, out_channels=complexity,
                                                  stride=(2, 2, 2), kernel_size=(4, 4, 4), padding=(1, 1, 1))
        self.batch_norm_transpose = nn.BatchNorm3d(complexity)

        self.conv3x3_2 = nn.Conv3d(complexity, complexity, kernel_size=3, padding=1)
        self.bn_2 = nn.BatchNorm3d(complexity)

        self.out_conv = nn.Conv3d(complexity, out_channels=out_channels, kernel_size=1, padding=0)

        self.activation = nn.LeakyReLU()

    def forward(self, x: torch.Tensor, i: int = 5) -> torch.Tensor:

        x = self.activation(self.bn_1(self.conv3x3_1(x)))
        x = self.activation(self.bn_5x5_1(self.conv5x5_1(x)))
        x = self.activation(self.bn_strided(self.strided_conv(x)))
        y = torch.zeros(x.shape).to(x.device)

        for t in range(i):
            in_ = torch.cat((x, y), dim=1)
            y = self.hcblock(in_) + y

        y = self.activation(self.batch_norm_transpose(self.transposed_conv(y)))
        y = self.activation(self.bn_2(self.conv3x3_2(y)))
        y = self.out_conv(y)

        return y


import torch
import hcat.lib.functional
from hcat.lib.functional import IntensityCellReject
from hcat.backends.backend import Backend
from hcat.models.r_unet import embed_model as RUnet
from hcat.train.transforms import median_filter, erosion
import hcat.lib.utils
from hcat.lib.utils import graceful_exit

import os.path
import wget

from typing import Dict, Optional


class SpatialEmbedding(Backend):
    def __init__(self,
                 sigma: Optional[torch.Tensor] = torch.tensor([0.02, 0.02, 0.02]),
                 device: Optional[str] = 'cuda',
                 model_loc: Optional[str] = None,
                 postprocessing: Optional[bool] = True,
                 scale: Optional[int] = 25,
                 figure: Optional[str] = None,
                 archetecture: Optional[RUnet] = RUnet):
        """
        Initialize Spatial embedding Algorithm.

        :param sigma: torch.Tensor[sigma_x, sigma_y, sigma_z] values for gaussian probability estimation.
        :param device: String value for torch device by which to run segmentation backbone on.
        :param model_loc: Path to trained model files.
        :param postprocessing: Disable segmentation postprocessing, namely
        :param scale: scale factor based on max diameter of object
        :param figure: filename and path of diagnostic figure which may be rendered
        """

        super(SpatialEmbedding, self).__init__()

        self.url = 'https://github.com/buswinka/hcat/blob/master/modelfiles/spatial_embedding.trch?raw=true'
        # self.url = None
        self.scale = torch.tensor(scale)
        self.device = device
        self.sigma = sigma.to(device)
        self.postprocessing = postprocessing

        self.figure = figure

        if self.url:
            self.model = self._model_loader_url(self.url, archetecture, device)
        else:
            self.model = self._model_loader_path(model_loc, archetecture, device)

        self.vector_to_embedding = torch.jit.script(
            hcat.lib.functional.VectorToEmbedding(scale=self.scale).requires_grad_(False).eval())

        self.embedding_to_probability = torch.jit.script(
            hcat.lib.functional.EmbeddingToProbability(scale=self.scale).requires_grad_(False).eval())

        self.estimate_centroids = hcat.lib.functional.EstimateCentroids(scale=self.scale).requires_grad_(False)

        self.filter = median_filter(kernel_targets=3, rate=1, device=device)
        self.binary_erosion = erosion(device=device)

        self.intensity_rejection = IntensityCellReject()

        self.nms = hcat.lib.functional.nms().requires_grad_(False)


        self.centroids = None
        self.vec = None
        self.embed = None
        self.prob = None

    @graceful_exit('\x1b[1;31;40m' + 'ERROR: Spatial Embedding Failed. Aborting...' + '\x1b[0m')
    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """
        Inputs an image and outputs a probability mask of everything seen in the image.

        .. note::
           Call the module as a function to execute this method (similar to torch.nn.module).

        .. warning:
           Will not raise an error upon failure, instead returns None and prints to standard out

        Example:

        >>> from hcat.backends.spatial_embedding import SpatialEmbedding
        >>> import torch
        >>> backend = SpatialEmbedding()
        >>> image = torch.load('path/to/my/image.trch')
        >>> assert image.ndim == 5 # Shape should be [B, C, X, Y, Z]
        >>> masks = backend(image)

        :param image: [B, C=4, X, Y, Z] input image
        :return: [B, 1, X, Y, Z] output segmentation mask where each pixel value is a cell id (0 is background)
        """
        assert image.ndim == 5
        assert image.shape[1] == 1
        assert image.min() >= -1
        assert image.max() <= 1

        # image = self.filter(image.to(self.device))
        image = image.to(self.device)
        b, c, x, y, z = image.shape

        if self.image_reject and self._is_image_bad(image):
            return torch.zeros((b, 0, x, y, z), device=self.device)


        # Evaluate Neural Network Model

        out: torch.Tensor = self.model(image)

        # Assign Outputs
        probability_map = out[:, [-1], ...]
        out = out[:, 0:3:1, ...]

        self.prob = probability_map.cpu()
        self.vec = out.cpu()

        out: torch.Tensor = self.vector_to_embedding(out)

        self.embed = out.cpu()

        centroids: Dict[str, torch.Tensor] = self.estimate_centroids(out, probability_map)

        self.centroids = centroids




        out: torch.Tensor = self.embedding_to_probability(out, centroids, self.sigma)


        # Reject cell masks that overlap or meet min Myo7a criteria
        if self.postprocessing:
            out: torch.Tensor = self.intensity_rejection(out, image)

        # print(centroids.shape, out.shape)

        if out.numel() == 0:
            return torch.zeros((b, 0, x, y, z), device=self.device)

        ind = self.nms(out, 0.5)
        out = out[:, ind, ...]

        # Take probabilities and generate masks!
        probability_map = probability_map.lt(0.8).squeeze(1)
        for i in range(out.shape[1]):
            out[:, i, ...][probability_map] = 0

        self.zero_grad()

        return out

    def load(self, model_loc: str) -> None:
        """
        Initializes model weights from a url or filepath.

        Example:

        >>> from hcat.backends.spatial_embedding import SpatialEmbedding
        >>> backend = SpatialEmbedding()
        >>>
        >>> url = 'https://www.model_location.com/model.trch'
        >>> backend.load(url) # Works with url
        >>>
        >>> model_path = 'path/to/my/model.trch'
        >>> backend.load(model_path) # Also works with path


        :param model_loc: url or filepath
        :return: None
        """
        if self._is_url(model_loc):
            return self._model_loader_url(model_loc, RUnet(in_channels=1).requires_grad_(False), self.device)
        else:
            return self._model_loader_path(model_loc, RUnet(in_channels=1).requires_grad_(False), self.device)





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    description="A Hair Cell Analysis Toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires='>=3',
    entry_points = {'console_scripts': ['hcat-segment=hcat.main:segment',
                                        'hcat-detect=hcat.main:detect',
                                        'hcat=hcat.main:cli']},
    install_requires = [
        'kornia>=0.5.2',
        'numpy>=1.10.0',
        'torch>=1.9.0',
        'matplotlib>=3.3.2',
        'scipy>=1.5.4',
        'scikit-image>=0.17.2',
        'tqdm>=4.60.0',
        'click>=8.0.0',
        'lz4>=3.1.3',
        'scikit-learn>=0.24.2',
        'GPy>=1.10.0',
        'torchvision>=0.10.0',
        'elasticdeform>=0.4.9',
        'wget>=3.2']
)